/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File dialog.c : User interface for SimCoupe
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"


/* Special Key codes */

#define KEY_NONE	0
#define KEY_UP		1
#define KEY_DOWN	2
#define KEY_LEFT	3
#define KEY_RIGHT	4
#define KEY_CTRL	5
#define KEY_ENTER	6

/* Drawing Modes */

#define DM_OVER	0
#define DM_AND	1
#define DM_OR	2
#define DM_XOR	3

/* Shadow Modes */

#define SHADOW_OUT	0
#define SHADOW_IN	1

#define SH_WID		2

/* Clut offsets for special colours */

#define SHCOL_LIGHT	13
#define SHCOL_MEDIUM	14
#define SHCOL_DARK	15

/* Character size constants */

#define GLYPH_WIDTH	6
#define GLYPH_HEIGHT	8

#define NO_GLYPHS	96

#define FONT_FILE	"font.bin"


/* Dialog Objects */

/* Object structure definitions */

#define MENU_OBJECT		0

#define MAX_MENU_ITEMS		25

#define MENU_DIALOG_ITEM	0
#define MENU_FUNCTION_ITEM	1
#define MENU_SUBMENU_ITEM	2
#define MENU_LINE_ITEM		3	

typedef struct _MenuObject {

	int NoMenuItems;
	int CurrentItem;
	struct _MenuItem {
	
		int   ItemType;
		char *ItemText;		
		union _ItemUnion {
			void *ItemPtr;
			void (*ItemFunction)(void);
		} ItemUnion;		

	} MenuItems[MAX_MENU_ITEMS];	

} MenuObject;


/* The main Dialog Structure */

typedef struct _DialogObject {

/* These items are common to all dialog object types */

	int	 ObjectType;
	uns8	 xpos, ypos;
	uns8	 width, height;
	uns8	*BackingStore;
	struct  _DialogObject *ParentObject;

	int	 IsMapped;

/* The rest of the Object is a Union which depends on the different Object types */

	union _ObjectUnion {
	
		MenuObject	AMenu;
			
	} ObjectUnion;
} DialogObject;

/* Public function prototypes */

void MapDialog(DialogObject *TheDialog);
void UnMapDialog(DialogObject *TheDialog);
DialogObject *MenuCreate(uns8 x, uns8 y, DialogObject *TheParent);
void MenuFree(DialogObject * TheDialog);
int MenuAddItem(DialogObject *TheDialog, int TheType, char *TheText, void* TheItemPtr);
void MenuDraw(DialogObject *TheDialog);
void MenuHighLight(DialogObject *TheDialog, int DoHighLight);
void MenuSelectItem(DialogObject *TheDialog, int TheItem);
void MenuCore(DialogObject *TheDialog);


int GetSpecialKey(void);

void DrawLineHorz(uns8 x, uns8 y, uns8 Length, uns8 Colour, int DrawMode);
void DrawLineVert(uns8 x, uns8 y, uns8 Length, uns8 Colour, int DrawMode);
void FillArea(uns8, uns8, uns8 width, uns8 height, uns8 Colour, int DrawMode);
void DrawString(uns8 x, uns8 y, char *TheString, uns8 Colour, int DrawMode);
void DrawShadowBox(uns8 x, uns8 y, uns8 width, uns8 height, int ShadowMode, int FillBox, int DrawMode);
void DrawBox(uns8 x, uns8 y, uns8 width, uns8 height, int Colour, int DrawMode);
void FetchArea(uns8 *Buffer, uns8 x, uns8 y, uns8 width, uns8 height);
void StoreArea(uns8 *Buffer, uns8 x, uns8 y, uns8 width, uns8 height);


void DialogInitFont(void);

void call_lmpr(uns8 lmpr_val);
void call_vmpr(uns8 vmpr_val);

uns8 *SamMemoryPages[2];	/* Pointers used to save the real memory page */
				/* addresses while the option screen is used  */
uns8 SamVmpr;			/* Saved value of vmpr                        */
uns8 SamLmpr;			/* Saved value of lmpr			      */
uns8 SamClut[16];		/* Saved clut registers			      */

#define SCREEN_LOC	30	/* Location in realMem[] of the option screen */

uns8 *DialogPages[2];

int DialogActive;
uns8 * DialogFontPtr;


static DialogObject *ActiveDialogObject = NULL;		/* The Active Dialog */

void DialogInit(void)
{

	int i;

	/* Allocate the two 16K pages required for the options screen */
	for (i = 0; i < 2; i++) {
		if ((DialogPages[i] = (uns8 *)malloc(SAM_PAGESIZE)) == (uns8 *)0)
			simcoupe_mesg(SIMCOUPE_FATAL, "can't allocate 16K for dialogs %d!", i);
	}
	
	DialogActive = False;

	/* Create the Dialog font */
	
	DialogInitFont();

}

DialogObject *TestDialog, *SubTestDialog;

/* Page in the dialog screen, after saving all the revevant Sam memory and ports */

void DialogStart(void)
{

	extern void DialogEnd(void);
	
	int i;

	InterruptType = Z80_halt;
	DialogActive = True;
	
	SamMemoryPages[0] = realMem[SCREEN_LOC+0];
	SamMemoryPages[1] = realMem[SCREEN_LOC+1];
	
	SamVmpr = vmpr;
	SamLmpr = lmpr;
	memcpy(SamClut, clut, 16);

	/* Move this when I use some more of the palette ! */

	clut[SHCOL_LIGHT]  = 119; 
	clut[SHCOL_MEDIUM] = 112;
	clut[SHCOL_DARK]   = 15;
		
	realMem[SCREEN_LOC+0] = DialogPages[0];
	realMem[SCREEN_LOC+1] = DialogPages[1];

	/* Page in the dialog screen in Section AB */

	vmpr = SCREEN_LOC | MODE_4 | 0x80;
	lmpr = SCREEN_LOC | 0x20;

	/* Do the actual paging */
	call_lmpr(lmpr);
	call_vmpr(vmpr);

	/* Create the Dialog font */
	
	DialogInitFont();

	TestDialog = MenuCreate(10,10, NULL);
	SubTestDialog = MenuCreate(50,10, TestDialog);

	MenuAddItem(TestDialog, MENU_SUBMENU_ITEM, "Menu", SubTestDialog);
	MenuAddItem(TestDialog, MENU_DIALOG_ITEM, "Item 2", NULL);
	MenuAddItem(TestDialog, MENU_DIALOG_ITEM, "Item 3", NULL);
	MenuAddItem(TestDialog, MENU_DIALOG_ITEM, "Item 4", NULL);
	MenuAddItem(TestDialog, MENU_FUNCTION_ITEM, "Quit", DialogEnd);
	
	MenuAddItem(SubTestDialog, MENU_DIALOG_ITEM, "Sub Item 1", NULL);
	MenuAddItem(SubTestDialog, MENU_DIALOG_ITEM, "Sub Item 2", NULL);
	MenuAddItem(SubTestDialog, MENU_DIALOG_ITEM, "Sub Item 3", NULL);
	MenuAddItem(SubTestDialog, MENU_DIALOG_ITEM, "Sub Item 4", NULL);


	ActiveDialogObject = TestDialog;

}

/* Page out the dialog screen, reseting all the revevant Sam memory and ports */

void DialogEnd(void)
{

	int i;
	
	realMem[SCREEN_LOC+0] = SamMemoryPages[0];
	realMem[SCREEN_LOC+1] = SamMemoryPages[1];
	
	vmpr = SamVmpr;
	lmpr = SamLmpr;
	memcpy(clut, SamClut, 16);

	/* Do the actual paging */
	call_lmpr(lmpr);
	call_vmpr(vmpr);

	DialogActive = False;
	InterruptType = Z80_restart;
	
}


DialogObject *CreateDialogObject(int TheType, uns8 x, uns8 y, uns8 TheWidth, uns8 TheHeight, 
		   DialogObject *TheParent);
void FreeDialogObject(DialogObject *TheDialog);


/* Dialog Creation and Freeing */

DialogObject *
CreateDialogObject(int TheType, uns8 x, uns8 y, uns8 TheWidth, uns8 TheHeight, 
		   DialogObject *TheParent)
{

	DialogObject *DialogPtr;

	/* Allocate the memory for a Dialog Object and fill in the position */
	/* and size 							    */

	DialogPtr = (DialogObject *)malloc(sizeof(DialogObject));
	
	DialogPtr -> ObjectType = TheType;
	DialogPtr -> xpos   = x;
	DialogPtr -> ypos   = y;
	DialogPtr -> width  = TheWidth;
	DialogPtr -> height = TheHeight;
	DialogPtr -> ParentObject = TheParent;

	DialogPtr -> IsMapped = False;
	DialogPtr -> BackingStore = NULL;	

	return (DialogPtr);

}

void
FreeDialogObject(DialogObject *TheDialog)  
{
	/* Free up the memory Occupied by TheDialog (including its BackingStore if */
	/* it is mapped								  */

	if (TheDialog -> IsMapped)
		free(TheDialog -> BackingStore);

	free(TheDialog);
	
}		  

void 
MapDialog(DialogObject *TheDialog)
{
	/* Allocate the backing store for TheDialog, if required, fetch    */
	/* the area of the display to be used and signal that it is mapped */
	
	size_t ObjectSize = (TheDialog->height)*((TheDialog->width+1)>>1);
	
	if (TheDialog -> BackingStore == NULL)
		TheDialog -> BackingStore = (uns8 *)malloc(ObjectSize);

	FetchArea(TheDialog -> BackingStore, TheDialog->xpos,   TheDialog->ypos, 
		  TheDialog->width, TheDialog->height);
	TheDialog -> IsMapped = True;
}

void
UnMapDialog(DialogObject *TheDialog)
{

	/* Unmap the dialog object and replace the display area covered */
	StoreArea(TheDialog -> BackingStore, TheDialog->xpos,   TheDialog->ypos, 
		  TheDialog->width, TheDialog->height);
	TheDialog -> IsMapped = False;
}

void
DialogCore(void)
{
	/* The core Dialog function, called from check_event() */
	
	/* Call the appropriate Dialog Object core code */
	
	if (ActiveDialogObject == NULL) return;

	switch (ActiveDialogObject -> ObjectType) {
	
		case MENU_OBJECT:
			MenuCore(ActiveDialogObject);
			break;	
	}

}


/* Menu Object routines */

DialogObject *
MenuCreate(uns8 x, uns8 y, DialogObject *TheParent)
{
	/* Create a Menu. NB The actual size of a menu is determined   */
	/* dynamically when items are added			       */
	
	DialogObject *DialogPtr;
	
	DialogPtr = CreateDialogObject(MENU_OBJECT, x, y, 2*SH_WID, 2*SH_WID, TheParent);
	
	DialogPtr -> ObjectUnion.AMenu.NoMenuItems = 0;
	DialogPtr -> ObjectUnion.AMenu.CurrentItem = 0;
	return(DialogPtr);
	
}

void
MenuFree(DialogObject * TheDialog)
{
	/* Free up all the memory used by a menu, and the Free the Object itself */
	
	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	int Item;
	
	for (Item = 0; Item < MenuPtr -> NoMenuItems; Item++) {
		free(MenuPtr -> MenuItems[Item].ItemText);
	}
	FreeDialogObject(TheDialog);
}

int
MenuAddItem(DialogObject *TheDialog, int TheType, char *TheText, void* TheItemPtr)
{
	/* Add an item to the menu, adjusting the size appropriately */

	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	
	if (MenuPtr->NoMenuItems == MAX_MENU_ITEMS)
		return(False);

	
	MenuPtr->MenuItems[MenuPtr->NoMenuItems].ItemType = TheType;
	MenuPtr->MenuItems[MenuPtr->NoMenuItems].ItemText = (char *)malloc(strlen(TheText));
	strcpy(MenuPtr->MenuItems[MenuPtr->NoMenuItems].ItemText, TheText);
	MenuPtr->MenuItems[MenuPtr->NoMenuItems].ItemUnion.ItemPtr  = TheItemPtr;	

	(TheDialog -> height) += GLYPH_HEIGHT + 2*SH_WID;
	
	if (strlen(TheText)*GLYPH_WIDTH > TheDialog -> width)
		 TheDialog -> width = 10 + strlen(TheText)*GLYPH_WIDTH;

	(MenuPtr->NoMenuItems)++;

	return(True);
}

void
MenuDraw(DialogObject *TheDialog)
{
	uns8 ItemX, ItemY;
	int  Item;

	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);

	/* Draw the Menu using the appropriate primatives   	      */
	/* The Object should be mapped before this function is called */

	/* First we draw a shadow box */
	
	DrawShadowBox(TheDialog->xpos,  TheDialog->ypos,
		      TheDialog->width, TheDialog->height,
		      SHADOW_OUT, True, DM_OVER);

	/* Now write in the menu entries */

	ItemX = TheDialog->xpos+2*SH_WID;
	ItemY = TheDialog->ypos+2*SH_WID;	
	for (Item = 0; Item < MenuPtr->NoMenuItems; Item++) {
		DrawString(ItemX, ItemY, MenuPtr->MenuItems[Item].ItemText, SHCOL_DARK, 
			   DM_OR);
		ItemY += GLYPH_HEIGHT + 2*SH_WID;
		
	}	
}

void
MenuHighLight(DialogObject *TheDialog, int DoHighLight)
{
	/* Highlight or unhighlight a given menu option */
	
	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	uns8 ItemX = TheDialog->xpos+ SH_WID;
	uns8 ItemY = TheDialog->ypos+ SH_WID+ MenuPtr->CurrentItem*(GLYPH_HEIGHT + 2*SH_WID);	


	if (!DoHighLight)
		DrawBox(ItemX, ItemY, 
			TheDialog->width - 2*SH_WID, GLYPH_HEIGHT+2*SH_WID,
			SHCOL_MEDIUM, DM_OVER);
	else
		DrawShadowBox(ItemX, ItemY,
		      TheDialog->width - 2*SH_WID, GLYPH_HEIGHT+2*SH_WID,
		      SHADOW_OUT, False, DM_OVER);
 

}

void
MenuSelectItem(DialogObject *TheDialog, int TheItem)
{
	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	
	/* Process a Menu selection */

	if (TheItem > MenuPtr->NoMenuItems ||
	    (MenuPtr -> MenuItems[TheItem].ItemUnion.ItemPtr == NULL)) return;

	switch (MenuPtr -> MenuItems[TheItem].ItemType) {
	
		case MENU_DIALOG_ITEM: case MENU_SUBMENU_ITEM:
			ActiveDialogObject = (DialogObject *)MenuPtr->MenuItems[TheItem].ItemUnion.ItemPtr;
			break;
			
		case MENU_FUNCTION_ITEM:
			MenuPtr->MenuItems[TheItem].ItemUnion.ItemFunction();
			break;	
	
	}
}

void
MenuCore(DialogObject *TheDialog)
{
	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	int KeyEvent;

	/* The core menu code, called by DialogCore */
	
	/* Map and draw the menu if required */
	
	if (!TheDialog -> IsMapped) {
		MapDialog(TheDialog);
		MenuDraw(TheDialog);
		MenuHighLight(TheDialog, True);
	}

	/* Deal with Events */
	
/*	printf("CurrentItem = %d\n",MenuPtr->CurrentItem);*/
	switch (KeyEvent=GetSpecialKey()) {
	
		case KEY_UP:
			if (MenuPtr->CurrentItem) {
				MenuHighLight(TheDialog, False);
				MenuPtr->CurrentItem--;
				MenuHighLight(TheDialog, True);

			}
			break;
		case KEY_DOWN:
			if (MenuPtr->CurrentItem<MenuPtr->NoMenuItems-1) {
				MenuHighLight(TheDialog, False);
				MenuPtr->CurrentItem++;
				MenuHighLight(TheDialog, True);
			}
			break;	
		case KEY_CTRL:
			if (TheDialog->ParentObject != NULL) {
				ActiveDialogObject = TheDialog->ParentObject;
				UnMapDialog(TheDialog);
			}
			break;
		case KEY_LEFT: case KEY_RIGHT:
			if (MenuPtr->MenuItems[MenuPtr->CurrentItem].ItemType == MENU_SUBMENU_ITEM)
				MenuSelectItem(TheDialog, MenuPtr->CurrentItem);
			else if (TheDialog->ParentObject != NULL) {
				ActiveDialogObject = TheDialog->ParentObject;
				UnMapDialog(TheDialog);
			}
			break;

		case KEY_ENTER:
			MenuSelectItem(TheDialog, MenuPtr->CurrentItem);
			break;
	}

}

/* Keyboard decoding functions */

int
GetSpecialKey(void)
{
	/* Return The status of the 'special' keys (cursors and CTRL) */

	uns8 Kstatus;
	int KeyCode = KEY_NONE;

/*	printf ("Kstatus = %d\n",Kstatus);*/
	Kstatus = (in_byte(KEYBOARD_PORT | 0xff00) & 0x1f);

	if (!(Kstatus & 0x02))
		KeyCode = KEY_UP;
	else if (!(Kstatus & 0x04))
		KeyCode = KEY_DOWN;
	else if (!(Kstatus & 0x08))
		KeyCode = KEY_LEFT;
	else if (!(Kstatus & 0x10))
		KeyCode = KEY_RIGHT;
	else if (!(Kstatus & 0x01))
		KeyCode = KEY_CTRL;
	
	if (KeyCode == KEY_NONE) {
		Kstatus = (in_byte(KEYBOARD_PORT | 0xbf00) & 0x1f);
		if (!(Kstatus & 0x01))
			KeyCode = KEY_ENTER;
	}

/*	printf ("KeyCode = %d\n",KeyCode);*/

	return (KeyCode);
}



/* Drawing Macros */

#define BytePos(x,y)	(((uns16)(y) << 7) + ((x) >> 1))
#define RightNybble(x)	((x) & 0x01)

#define R_MASK		0x0f
#define L_MASK		0xf0

#define FontOffset(c)	((c-32)*(GLYPH_HEIGHT*(GLYPH_WIDTH >> 1)))

/* Dialog primitives */

void
DialogInitFont(void)
{

	FILE *ifp;
	size_t nread, FontSize = NO_GLYPHS*GLYPH_HEIGHT*(GLYPH_WIDTH >> 1);

	/* Allocate memory for the font data */
	
	DialogFontPtr = (uns8 *)malloc(FontSize);
	
	/* Read in the font data */
	
	if ((ifp = open_simcoupe_file(FONT_FILE, "r")) == NULL) {
		simcoupe_mesg(SIMCOUPE_ERR, "can't open Dialog font file  <%s>", FONT_FILE);
		return;
	}

	nread = fread(DialogFontPtr, 1, FontSize, ifp);

	if (nread != FontSize ) {
		simcoupe_mesg(SIMCOUPE_WARN, "Font data file  <%s> is only %d bytes", FONT_FILE, nread);
	}

	fclose(ifp);

}

static inline void
PlotLeft(uns16 Addr, uns8 Colour, int DrawMode)
{
	uns8 OldByte = read_byte(Addr);
	
	switch (DrawMode) {
	
		case DM_OVER:
			write_byte(Addr,(OldByte & R_MASK) | ((Colour & R_MASK) <<4));
			break;
		case DM_AND:
			write_byte(Addr,(OldByte & R_MASK) | ((OldByte & L_MASK) & ((Colour & R_MASK) <<4)));
			break;
		case DM_OR:
			write_byte(Addr,(OldByte & R_MASK) | ((OldByte & L_MASK) | ((Colour & R_MASK) <<4)));
			break;			
		case DM_XOR:
			write_byte(Addr,(OldByte & R_MASK) | ((OldByte & L_MASK) ^ ((Colour & R_MASK) <<4)));
			break;
	}

}
static inline void
PlotRight(uns16 Addr, uns8 Colour, int DrawMode)
{
	uns8 OldByte = read_byte(Addr);
	
	switch (DrawMode) {
	
		case DM_OVER:
			write_byte(Addr,(OldByte & L_MASK) | (Colour & R_MASK));
			break;
		case DM_AND:
			write_byte(Addr,(OldByte & L_MASK) | ((OldByte & R_MASK) & (Colour & R_MASK)));
			break;
		case DM_OR:
			write_byte(Addr,(OldByte & L_MASK) | ((OldByte & R_MASK) | (Colour & R_MASK)));
			break;			
		case DM_XOR:
			write_byte(Addr,(OldByte & L_MASK) | ((OldByte & R_MASK) ^ (Colour & R_MASK)));
			break;
	}

}

static inline void
PlotBoth(uns16 Addr, uns8 Colours, int DrawMode)
{
	uns8 OldByte = read_byte(Addr);
	uns8 NewByte = Colours;
	switch (DrawMode) {
	
		case DM_OVER:
			write_byte(Addr,NewByte);
			break;
		case DM_AND:
			write_byte(Addr,OldByte & NewByte);
			break;
		case DM_OR:
			write_byte(Addr,OldByte | NewByte);
			break;			
		case DM_XOR:
			write_byte(Addr,OldByte ^ NewByte);
			break;
	}

}


void 
DrawLineHorz(uns8 x, uns8 y, uns8 Length, uns8 Colour, int DrawMode)
{

	uns16 ScrnByte = BytePos(x,y);
	int LengthLeft = Length;
	
	if (RightNybble(x)) {
		PlotRight(ScrnByte++, Colour, DrawMode);		
		LengthLeft--;
	}

	while (LengthLeft > 1) {
		PlotBoth(ScrnByte++,((Colour & R_MASK) << 4 | (Colour & R_MASK)), DrawMode);
		LengthLeft -=2;
	}

	if (LengthLeft !=0 ) {
		PlotLeft(ScrnByte, Colour, DrawMode);
	}

}

void 
DrawLineVert(uns8 x, uns8 y, uns8 Length, uns8 Colour, int DrawMode)
{

	uns16 ScrnByte = BytePos(x,y);
	int LengthLeft = Length;	

	for (; LengthLeft--; ScrnByte+=128) {
	
		if (RightNybble(x))
			PlotRight(ScrnByte, Colour, DrawMode);
		else
			PlotLeft(ScrnByte, Colour, DrawMode); 
	}
}

void
FillArea(uns8 x, uns8 y, uns8 width, uns8 height, uns8 Colour, int DrawMode)
{
	int i;
	
	for (i=0; i<height; i++)
		DrawLineHorz(x, y++, width, Colour, DrawMode);

}

void
DrawChar(uns8 x, uns8 y, char Character, uns8 Colour, int DrawMode)
{
	int i;
	uns16 ScrnByte = BytePos(x,y);
	uns8 *GlyphPtr = DialogFontPtr + FontOffset(Character)+1;
	
	/* Fill both Nybbles with Colour */

	Colour = ((Colour & R_MASK) << 4 | (Colour & R_MASK));

	for (i = 0; i < GLYPH_HEIGHT; i++) {
		PlotBoth(ScrnByte++, Colour & *(GlyphPtr  ), DrawMode);
		PlotBoth(ScrnByte++, Colour & *(GlyphPtr+1), DrawMode);
		PlotBoth(ScrnByte++, Colour & *(GlyphPtr+2), DrawMode);
		GlyphPtr+=3;
		ScrnByte+=125;
	}
}

void
DrawString(uns8 x, uns8 y, char *TheString, uns8 Colour, int DrawMode)
{

	int  i;
	
	for (i = 0; i < strlen(TheString); i++) {
	
		DrawChar(x, y, TheString[i], Colour, DrawMode);
		x+=GLYPH_WIDTH;
	}

}

void
DrawShadowBox(uns8 x, uns8 y, uns8 width, uns8 height, int ShadowMode, int FillBox, int DrawMode)
{

	uns8 TLColour, BRColour;

	/* Sort out the Colours to use */

	switch(ShadowMode) {
	
		case SHADOW_OUT:
			TLColour = SHCOL_LIGHT;
			BRColour = SHCOL_DARK;
			break;
		case SHADOW_IN:
			TLColour = SHCOL_DARK;
			BRColour = SHCOL_LIGHT;
			break;
	}

	/* Now draw the borders of the box */
	
	DrawLineHorz(x, y, width,      TLColour, DrawMode);
	DrawLineHorz(x, y+1, width-1,  TLColour, DrawMode);
	DrawLineVert(x, y, height,     TLColour, DrawMode);
	DrawLineVert(x+1, y, height-1, TLColour, DrawMode);	

	DrawLineHorz(x+1, y+height-1, width-1, BRColour, DrawMode);
	DrawLineHorz(x+2, y+height-2, width-2, BRColour, DrawMode);
	DrawLineVert(x+width-1, y+1, height-1, BRColour, DrawMode);
	DrawLineVert(x+width-2, y+2, height-2, BRColour, DrawMode);

	/* Fill the box if required */
	
	if (FillBox)
		FillArea(x+2,y+2,width-4, height-4, SHCOL_MEDIUM, DrawMode);

}

void
DrawBox(uns8 x, uns8 y, uns8 width, uns8 height, int Colour, int DrawMode)
{
	
	DrawLineHorz(x, y, width,      Colour, DrawMode);
	DrawLineHorz(x, y+1, width-1,  Colour, DrawMode);
	DrawLineVert(x, y, height,     Colour, DrawMode);
	DrawLineVert(x+1, y, height-1, Colour, DrawMode);	

	DrawLineHorz(x+1, y+height-1, width-1, Colour, DrawMode);
	DrawLineHorz(x+2, y+height-2, width-2, Colour, DrawMode);
	DrawLineVert(x+width-1, y+1, height-1, Colour, DrawMode);
	DrawLineVert(x+width-2, y+2, height-2, Colour, DrawMode);

}

void
FetchArea(uns8 *Buffer, uns8 x, uns8 y, uns8 width, uns8 height)
{
	/* Fetch the required display area and store in the (prealloced) */
	/* Memory pointed to by Buffer					 */

	uns16 ScrnByte = BytePos(x,y);
	int ByteWidth = BytePos(x+width+1,y) - ScrnByte;
	int i;
	
	while (height--) {
		for (i=0; i < ByteWidth; i++)
			*Buffer++ = read_byte(ScrnByte++);
		ScrnByte+= (128-ByteWidth);
	}
}



void
StoreArea(uns8 *Buffer, uns8 x, uns8 y, uns8 width, uns8 height)
{

	/* Store  the required display area from the (preallocated) */
	/* Memory pointed to by Buffer				    */

	uns16 ScrnByte = BytePos(x,y);
	int ByteWidth = BytePos(x+width+1,y) - ScrnByte;
	int i;
	
	while (height--) {
		for (i=0; i < ByteWidth; i++)
			write_byte(ScrnByte++, *Buffer++);
/*			write_byte(ScrnByte++, 0xff); */
		ScrnByte+= (128-ByteWidth);
	}
}
