/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File interface.c - User interface implementation using libSCdialog 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"
#include "version.h"
#include "dialog.h"
#include "fdi.h"
#include "resource.h"
#include "intrface.h"

int UiActive;
int UiDone;

#define MAX_PATH	1024
char TheCwd[MAX_PATH];

/* UI Pointer status */

static int UIPointerDeltaX;
static int UIPointerDeltaY;
static int UIButtonState;

extern int disass;
extern int SAMPointerMode;

void
ui_init(void)
{
	/* User interface initialisation - called once at startup */

	DialogInit();

	/* Get the Current Working Directory - useful for file i/o */

	if (getcwd(TheCwd, (size_t)MAX_PATH) == NULL) {
#ifndef __DJGPP
		strcpy(TheCwd, "/");
	}
#else
		strcpy(TheCwd, "c:\\");
	}
	else {
		char *PathPtr=TheCwd;
		while (*PathPtr++) {
			if (*PathPtr == '/')
				*PathPtr = '\\';
		}
	}
#endif	

	ui_set_button_state(0x07);
	ui_set_pointer_position(0, 0);
	UiActive = False;
}

void
ui_end(void)
{

	/* User interface Closedown */

	DialogCloseDown();

}

void
ui_enter(void)
{
	/* Enter the UI from the emulation */

	/* Get the Current Working Directory - useful for file i/o */

	if (getcwd(TheCwd, (size_t)MAX_PATH) == NULL) {
#ifndef __DJGPP
		strcpy(TheCwd, "/");
	}
#else
		strcpy(TheCwd, "c:\\");
	}
	else {
		char *PathPtr=TheCwd;
		while (*PathPtr++) {
			if (*PathPtr == '/')
				*PathPtr = '\\';
		}
	}
#endif	
	/* Halt the processor at the next frame interrupt */

	InterruptType = Z80_halt;
	UiDone   = False;
	UiActive = True;

	DialogStart();	

}

void
ui_leave(void)
{
	/* Leave the UI and return to the emulation */

	ui_set_button_state(0x07);
	ui_set_pointer_position(0, 0);
	PointerSetStatus(0, 0, 0x7);
	DialogEnd();


	/* Restart the processor */

	UiActive = False;
	InterruptType = Z80_restart;

}

void
ui_set_button_state(int ButtonState )
{
	/* Set Mouse Button status for UI */
	
	UIButtonState = ButtonState & 0x07;

}

void
ui_set_pointer_position( int PointerDeltaX, int PointerDeltaY )
{

	/* Set Mouse Pointer position for UI */
	
	UIPointerDeltaX = PointerDeltaX;
	UIPointerDeltaY = PointerDeltaY;

}

void
ui_pointer_draw(void)
{
	PointerDraw();
}

void
ui_pointer_erase(void)
{
	PointerErase();
}

void
ui_event( void )
{
	/* Deal with UI events */


	if (SAMPointerMode) {
		PointerErase();

		PointerSetStatus(UIPointerDeltaX, UIPointerDeltaY, UIButtonState);
	}
	else {
		PointerSetStatus(-999, -999, -1);
	}

	if (DialogGetActiveDialogObject() == NULL) {
#if 0
		int KeyEvent = GetSpecialKey();
		if (KeyEvent != KEY_NONE)
			ui_leave();
#endif

	}
	else {
		DialogCore();
	}

	if (SAMPointerMode) {
		PointerDraw();
	}

	/* If we are done leave the user interface */
	if (UiDone)
		ui_leave();

}


DialogObject *
ui_make_about(void (*CBAbout)(void), DialogObject *AboutParent)
{
	/* Create an 'About' Dialog used both for the splash banner */
	/* and the main user interface 				    */

	DialogObject *TheAboutDialog;
	char buffer[255];
	
	sprintf(buffer, "SimCoupe vn %d.%d.%d",SIMCOUPE_Major,SIMCOUPE_Minor,SIMCOUPE_Patch);

	TheAboutDialog = FormCreate(50, 50, 156, 92, AboutParent);
	FormAddTextItem(TheAboutDialog, 28, 10, buffer);
	FormAddTextItem(TheAboutDialog, 28, 12, "_________________");
	FormAddTextItem(TheAboutDialog, 10, 30, "(c) 1997 A.J.Skillman");
		
	FormAddButtonItem(TheAboutDialog, 70,60, "OK", FORM_FUNCTION_BUTTON, CBAbout);
	
	return(TheAboutDialog);
}


static DialogObject *BannerDialog;

void
ui_banner(void)
{
	/* Use the UI to put up a welcoming banner */

	ui_enter();


	BannerDialog = ui_make_about(ui_end_banner, NULL);
	DialogSetActiveDialogObject(BannerDialog);
}

void 
ui_end_banner(void)
{
	DialogSetActiveDialogObject(NULL);
	FormFree(BannerDialog);
	ClearDisplay();
	UiDone = True;
}

static DialogObject *TopMenuForm;
static DialogObject *FileMenu;
static DialogObject *SamToolsMenu;
static DialogObject *OptionsMenu;
static DialogObject *HelpMenu;
static DialogObject *SelectDriveMenu;
static DialogObject *AboutDialog;
static DialogObject *CreateImageDialog;
static DialogObject *CreateImageExecDialog;
static DialogObject *FileSelectionRoot;
static DialogObject *FlashModeDialog;
static DialogObject *ExternalMemDialog;
static DialogObject *SynchronizeDialog;
static DialogObject *SpeakerDialog;

void
ui_start_debug(void)
{
	/* Start the SimCoupe debugger - just console mode for now */
	/* So exit the ui and go back to the normal emulation mode */
	
	disass = !disass;
	ui_main_end();	

}

void
ui_CBselect_drive1(char *TheDiskImageName)
{

	if (strlen(TheDiskImageName)) {
		eject_disk(1);
		insert_disk(TheDiskImageName, 1);
	}
}

void
ui_select_drive1(void)
{

	FileSelectorOpen(TheCwd, "*.dsk,*.sad", ui_CBselect_drive1, DialogGetActiveDialogObject());

}

void
ui_CBselect_drive2(char *TheDiskImageName)
{

	if (strlen(TheDiskImageName)) {
		eject_disk(2);
		insert_disk(TheDiskImageName, 2);
	}
}


void
ui_select_drive2(void)
{

	FileSelectorOpen(TheCwd, "*.dsk,*.sad", ui_CBselect_drive2, DialogGetActiveDialogObject());

}

void
ui_about_return(void)
{

	DialogSetActiveDialogToRoot();
	
}

/* Data associated with the Create disk image dialogs */
static char ImageName[BUFSIZ] = "Image.dsk";
static int 	ImageAddSystem;		 

/* Data associated with Options dialogs */
static int *Mode1Flash;
static int *Synchronize;
static int *Speaker;
static char NoMbStr[3] = "0";

void
ui_create_image(void)
{
    create_disk_image(ImageName, ImageAddSystem);
    ui_main_end();
}

void
ui_to_parent(void)
{
	DialogSetActiveDialogToParent();
}

void
ui_make_file(void)
{

	/* Create the Create Disk image dialog (File -> Create Image) */

    CreateImageExecDialog = ExecCreate(0, 0, 0, 0, NULL);
    ExecSetEventFunction(CreateImageExecDialog, ui_create_image);

	CreateImageDialog = FormCreate(20, 50, 215, 70, FileMenu);
	FormAddTextItem(CreateImageDialog, 6, 12, "File Name :");
	FormAddInputBoxItem(CreateImageDialog, 80, 8, BUFSIZ, 20, ImageName, NULL);
	FormAddTextItem(CreateImageDialog, 60, 30, "Make System Disk");
	FormAddSwitchItem(CreateImageDialog, 44, 32, 2, 2, &ImageAddSystem, True, False);
	FormAddButtonItem(CreateImageDialog, 20, 48, "  OK  ", FORM_DIALOG_BUTTON, CreateImageExecDialog);
	FormAddButtonItem(CreateImageDialog, 150, 48,"Cancel", FORM_FUNCTION_BUTTON, ui_to_parent);

	/* Create the drive selection sub-menu (File -> Change Disk) */
	SelectDriveMenu = MenuCreate(66, 2*GLYPH_HEIGHT+6*SH_WID, FileMenu);	
	MenuAddItem(SelectDriveMenu, MENU_FUNCTION_ITEM, "Disk Drive 1", ui_select_drive1);
	MenuAddItem(SelectDriveMenu, MENU_FUNCTION_ITEM, "Disk Drive 2", ui_select_drive2);


}


void
ui_make_options(void)
{
	Mode1Flash = &GetCfg(flashing);
	FlashModeDialog = FormCreate(40, 80, 175, 60, OptionsMenu);
	FormAddTextItem(FlashModeDialog, 35, 20, "Emulate Mode 1 Flash");
	FormAddSwitchItem(FlashModeDialog, 19, 22, 2, 2, Mode1Flash, True, False);
	FormAddButtonItem(FlashModeDialog, 65, 38, "  OK  ", FORM_FUNCTION_BUTTON, ui_to_parent);

	Synchronize = &GetCfg(sync);
	SynchronizeDialog = FormCreate(40, 80, 175, 60, OptionsMenu);
	FormAddTextItem(SynchronizeDialog, 35, 20, "50Hz Synchronization");
	FormAddSwitchItem(SynchronizeDialog, 19, 22, 2, 2, Synchronize, True, False);
	FormAddButtonItem(SynchronizeDialog, 65, 38, "  OK  ", FORM_FUNCTION_BUTTON, ui_to_parent);

	Speaker = &GetCfg(use_speaker);
	SpeakerDialog = FormCreate(40, 80, 175, 60, OptionsMenu);
	FormAddTextItem(SpeakerDialog, 35, 20, "Emulate SAM speaker");
	FormAddSwitchItem(SpeakerDialog, 19, 22, 2, 2, Speaker, True, False);
	FormAddButtonItem(SpeakerDialog, 65, 38, "  OK  ", FORM_FUNCTION_BUTTON, ui_to_parent);

	sprintf(NoMbStr,"%d",GetCfg(externalMem));
	ExternalMemDialog = FormCreate(20, 80, 215, 60, OptionsMenu);
	FormAddTextItem(ExternalMemDialog, 6, 12, "No External Mb :");
	FormAddInputBoxItem(ExternalMemDialog, 120, 8, 1, 1, NoMbStr, NULL);
	FormAddButtonItem(ExternalMemDialog, 20, 38, "  OK  ", FORM_DIALOG_BUTTON, NULL);
	FormAddButtonItem(ExternalMemDialog, 150, 38,"Cancel", FORM_DIALOG_BUTTON, NULL);

}

void
ui_make_file_menu(void)
{

    ui_make_file();

	MenuAddItem(FileMenu, MENU_SUBMENU_ITEM, "Change Disk", SelectDriveMenu);
	MenuAddItem(FileMenu, MENU_DIALOG_ITEM,  "Create Image ", CreateImageDialog);
/* No Debug option for vga display - yet */
#ifndef VGA_DISPLAY
	MenuAddItem(FileMenu, MENU_FUNCTION_ITEM, "Start Debug", ui_start_debug);
#endif
	MenuAddItem(FileMenu, MENU_FUNCTION_ITEM, "Exit", ui_main_end);

}

void
ui_make_samtools_menu(void)
{

	MenuAddItem(SamToolsMenu, MENU_DIALOG_ITEM, "SamDir", NULL);

}

void
ui_make_options_menu(void)
{
	ui_make_options();

	MenuAddItem(OptionsMenu, MENU_DIALOG_ITEM, "Flash mode", FlashModeDialog);
#if 0
	MenuAddItem(OptionsMenu, MENU_DIALOG_ITEM, "External Mem.", ExternalMemDialog );
#endif
	MenuAddItem(OptionsMenu, MENU_DIALOG_ITEM, "Synchronize ", SynchronizeDialog);
	MenuAddItem(OptionsMenu, MENU_DIALOG_ITEM, "Speaker", SpeakerDialog);

}

void
ui_make_help_menu(void)
{
	/* Create the About dialog (Help -> About)*/
	AboutDialog  = ui_make_about(ui_about_return, HelpMenu);
	MenuAddItem(HelpMenu, MENU_DIALOG_ITEM, "About", AboutDialog);
}

void
ui_main(int InterfaceEntry)
{
	/* The main user interface to SimCoupe */

	ui_enter();
	
/*	ClearDisplay();*/

	/* First Create the top level menus */
	
	TopMenuForm  = FormCreate(0, 0, 255, GLYPH_HEIGHT+8*SH_WID, NULL);
	FileMenu     = MenuCreate(4,   GLYPH_HEIGHT+6*SH_WID, TopMenuForm);	
	SamToolsMenu = MenuCreate(40,  GLYPH_HEIGHT+6*SH_WID, TopMenuForm);	
	OptionsMenu  = MenuCreate(100, GLYPH_HEIGHT+6*SH_WID, TopMenuForm);	
	HelpMenu     = MenuCreate(154, GLYPH_HEIGHT+6*SH_WID, TopMenuForm);	

    /* Set up the main menu */
	FormAddButtonItem(TopMenuForm, 4,   4, "File", FORM_DIALOG_BUTTON, FileMenu);
	FormAddButtonItem(TopMenuForm, 40,  4, "SamTools", FORM_DIALOG_BUTTON, SamToolsMenu);
	FormAddButtonItem(TopMenuForm, 100, 4, "Options", FORM_DIALOG_BUTTON, OptionsMenu);
	FormAddButtonItem(TopMenuForm, 154, 4, "Help", FORM_DIALOG_BUTTON, HelpMenu);

	/* Set up the menus */
	
	ui_make_file_menu();
	ui_make_samtools_menu();
	ui_make_options_menu();
	ui_make_help_menu();

	FileSelectionRoot = ExecCreate(0, 0, 0, 0, NULL);
	ExecSetEventFunction(FileSelectionRoot, ui_main_end);

	switch(InterfaceEntry) {
	
	   case UI_MENU:
		DialogSetActiveDialogObject(TopMenuForm);
		break;

	   case UI_DISK1:
		DialogSetActiveDialogObject(FileSelectionRoot);		
        ui_select_drive1();
		break;

	   case UI_DISK2:
		DialogSetActiveDialogObject(FileSelectionRoot);		
        ui_select_drive2();
		break;

	}

}

void
ui_main_end(void)
{
	DialogSetActiveDialogObject(NULL);
	ExecFree(FileSelectionRoot);
	FormFree(TopMenuForm);
	MenuFree(FileMenu);
	MenuFree(SamToolsMenu);
	MenuFree(OptionsMenu);
	MenuFree(HelpMenu);

	FormFree(CreateImageDialog);
	FormFree(FlashModeDialog);
	
	ClearDisplay();
	UiDone = True;
}

