/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File dialog_core.c : User interface for SimCoupe : Core dialog functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"

#include "dialog.h"

void call_lmpr(uns8 lmpr_val);
void call_vmpr(uns8 vmpr_val);

uns8 *SamMemoryPages[2]; /* Pointers used to save the real memory page */
				         /* addresses while the option screen is used  */
uns8  SamVmpr;			 /* Saved value of vmpr                */
uns8  SamLmpr;			 /* Saved value of lmpr			      */
uns32 SamClut[16];		 /* Saved clut registers			      */
uns8  SamBorder;         /* Saved border value                 */

#define SCREEN_LOC	30	/* Location in realMem[] of the option screen */

uns8 *DialogPages[2];

uns8 		    *DialogFontPtr;
DialogObject	*ActiveDialogObject;		/* The Active Dialog */

void
DialogInit(void)
{

	int i;

	/* Allocate the two 16K pages required for the options screen */
	for (i = 0; i < 2; i++) {
		if ((DialogPages[i] = (uns8 *)malloc(SAM_PAGESIZE)) == (uns8 *)0)
			simcoupe_mesg(SIMCOUPE_FATAL, "can't allocate 16K for dialogs %d!", i);
	}
	
	/* Create the Dialog font */
	
	DialogInitFont();

	/* Initialise the pointer */

	PointerInit();
	
	ActiveDialogObject = NULL;
}

/* Page in the dialog screen, after saving all the revevant Sam memory and ports */

void
DialogStart(void)
{
	
	int i;

	SamMemoryPages[0] = realMem[SCREEN_LOC+0];
	SamMemoryPages[1] = realMem[SCREEN_LOC+1];
	
	SamVmpr = vmpr;
	SamLmpr = lmpr;
	memcpy(SamClut, clut, 16*4);

    SamBorder = border;

	/* Move this when I use some more of the palette ! */

	clut[SHCOL_LIGHT]  = 119; 
	clut[SHCOL_MEDIUM] = 112;
	clut[SHCOL_DARK]   = 15;
	clut[SHCOL_BLACK]  = 0;
	clut[SHCOL_WHITE]  = 127;		

    clut[USER_COL+0]   = 0x01;
    clut[USER_COL+1]   = 0x11;
    clut[USER_COL+2]   = 0x19;
    clut[USER_COL+3]   = 0x22; 
    clut[USER_COL+4]   = 0x54; 


	realMem[SCREEN_LOC+0] = DialogPages[0];
	realMem[SCREEN_LOC+1] = DialogPages[1];

	/* Page in the dialog screen in Section AB */

	vmpr = SCREEN_LOC | MODE_4 | 0x80;
	lmpr = SCREEN_LOC | 0x20;

	/* Do the actual paging */
	call_lmpr(lmpr);
	call_vmpr(vmpr);

	ActiveDialogObject = NULL;

    /* Draw the desktop */
    
    DesktopDraw();

	PointerEnter();
}

/* Page out the dialog screen, reseting all the revevant Sam memory and ports */

void
DialogEnd(void)
{
	int i;
	
	PointerExit();

	ClearDisplay();

	realMem[SCREEN_LOC+0] = SamMemoryPages[0];
	realMem[SCREEN_LOC+1] = SamMemoryPages[1];
	
	vmpr = SamVmpr;
	lmpr = SamLmpr;
	memcpy(clut, SamClut, 16*4);

    border = SamBorder;

	/* Do the actual paging */
	call_lmpr(lmpr);
	call_vmpr(vmpr);

}

void
DialogCloseDown(void)
{
	/* Closedown the dialog library */
	
	ActiveDialogObject = NULL;
	DialogFreeFont(); 
	PointerClose();

}


/* Dialog Creation and Freeing */

DialogObject *
CreateDialogObject(int TheType, uns8 x, uns8 y, uns8 TheWidth, uns8 TheHeight, 
		   DialogObject *TheParent)
{

	DialogObject *DialogPtr;

	/* Allocate the memory for a Dialog Object and fill in the position */
	/* and size 							    */

	DialogPtr = (DialogObject *)malloc(sizeof(DialogObject));
	
	DialogPtr -> ObjectType = TheType;
	DialogPtr -> xpos   = x;
	DialogPtr -> ypos   = y;
	DialogPtr -> width  = TheWidth;
	DialogPtr -> height = TheHeight;
	DialogPtr -> ParentObject = TheParent;

	DialogPtr -> IsMapped = False;
	DialogPtr -> BackingStore = NULL;	

	return (DialogPtr);

}

void
FreeDialogObject(DialogObject *TheDialog)  
{
	/* Free up the memory Occupied by TheDialog (including its BackingStore if */
	/* it is mapped								  */

	if (TheDialog -> IsMapped)
		free(TheDialog -> BackingStore);

	free(TheDialog);
	TheDialog = NULL;	
}		  

void 
MapDialog(DialogObject *TheDialog)
{
	/* Allocate the backing store for TheDialog, if required, fetch    */
	/* the area of the display to be used and signal that it is mapped */
	
	size_t ObjectSize = (TheDialog->height)*((TheDialog->width+2)>>1);
	
	if (TheDialog -> BackingStore == NULL)
		TheDialog -> BackingStore = (uns8 *)malloc(ObjectSize);

	FetchArea(TheDialog -> BackingStore, TheDialog->xpos,   TheDialog->ypos, 
		  TheDialog->width, TheDialog->height);
	TheDialog -> IsMapped = True;
}

void
UnMapDialog(DialogObject *TheDialog)
{

	/* Unmap the dialog object and replace the display area covered */
	StoreArea(TheDialog -> BackingStore, TheDialog->xpos,   TheDialog->ypos, 
		  TheDialog->width, TheDialog->height);
	TheDialog -> IsMapped = False;
}

void
DialogCore(void)
{
	/* The core Dialog function, called from the application main loop */
	
	/* Call the appropriate Dialog Object core code */
	
	if (ActiveDialogObject == NULL) return;

    /* Initialise the key state (once per cycle) */
    SetTabKeyState();
    SetSpecialKey();
	
	switch (ActiveDialogObject -> ObjectType) {

		case MENU_OBJECT:
			MenuCore(ActiveDialogObject);
			break;	
		case FORM_OBJECT:
			FormCore(ActiveDialogObject);
			break;	
		case EXEC_OBJECT:
			ExecCore(ActiveDialogObject);
			break;	
	}	
}

void
DialogInitFont(void)
{

	FILE *ifp;
	size_t nread, FontSize = NO_GLYPHS*GLYPH_HEIGHT*(GLYPH_WIDTH >> 1);

	/* Allocate memory for the font data */
	
	DialogFontPtr = (uns8 *)malloc(FontSize);
	
	/* Read in the font data */
	
	if ((ifp = open_simcoupe_file(FONT_FILE, "rb")) == NULL) {
		simcoupe_mesg(SIMCOUPE_ERR, "can't open Dialog font file  <%s>", FONT_FILE);
		return;
	}

	nread = fread(DialogFontPtr, 1, FontSize, ifp);

	if (nread != FontSize ) {
		simcoupe_mesg(SIMCOUPE_WARN, "Font data file  <%s> is only %d bytes", FONT_FILE, nread);
	}

	fclose(ifp);

}

void
DialogFreeFont(void)
{

	free(DialogFontPtr);

}

void
DialogSetActiveDialogObject(DialogObject *TheDialog)
{
	/* Set the Current Active Dialog */
	
	ActiveDialogObject = TheDialog;

}

DialogObject *
DialogGetActiveDialogObject(void)
{
	/* Get the Current Active Dialog */
	
	return(ActiveDialogObject);

}

void
DialogSetActiveDialogToParent(void)
{
	/* Go back up the tree one level (unless we are at the root) */
	
	if (ActiveDialogObject->ParentObject != NULL) {
	
		UnMapDialog(ActiveDialogObject);
		DialogSetActiveDialogObject(ActiveDialogObject->ParentObject);
	
	}
	else {
		UnMapDialog(ActiveDialogObject);
		DialogSetActiveDialogObject(NULL);

	}
}

void
DialogSetActiveDialogToRoot(void)
{
	/* Go back up the tree until we reach the root) */
	
	while (ActiveDialogObject->ParentObject != NULL) {
	
		UnMapDialog(ActiveDialogObject);
		DialogSetActiveDialogObject(ActiveDialogObject->ParentObject);
	
	}
}

