/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File file_selector.c : User interface for SimCoupe : File selector dialog
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* The FileSelector dialog is just a specialised use of the 'form' dialog,   */
/* which allows the user to browse through the file system and select a file.*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fnmatch.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"
#include "fdi.h"
#include "dialog.h"

/* String used to indicate Non applicable dir/file lists */
#define NA	"N/A"

/* Internal File Selector global variables */

#define MAX_PATH	BUFSIZ
#define MAX_DIRECTORIES	BUFSIZ
#define MAX_FILES	BUFSIZ
#define MAX_FILTER	255

/* Private types */

typedef struct _FileList {
	int NoFiles;
	char *FileNames[MAX_FILES];
} FileList;


/* Private function prototypes */

FileList *FileSelectorGetDirList(char *ThePath); 
FileList *FileSelectorGetFileList(char *ThePath, char *TheFilter);
void FileSelectorFreeFileList(FileList *TheFileList);

void CBFileSelectorOK(void);
void CBFileSelectorFilter(void);
void CBFileSelectorCancel(void);
void CBFileSelectorSelectDir(char *TheDir);
void CBFileSelectorSelectFile(char *TheFile);
void CBFileSelectorSelectFilePath(char *ThePath);
void CBFileSelectorSelectFilter(char *TheFilter);

static DialogObject *TheFileSelectorDialog = NULL;
static FormItem *DirListBox;
static FormItem *FileListBox;
static FormItem *PathInputBox;
static FormItem *FilterInputBox;
static char PathBuffer[MAX_PATH];
static char FilterBuffer[MAX_FILTER];


static char CurrentPath[MAX_PATH];
static char CurrentFilter[MAX_FILTER];

static char SelectedFile[MAX_PATH];

static FileList *CurrentDirectories;
static FileList *CurrentFiles;


void (*CBuser)(char *);


void
FileSelectorOpen(char *ThePath, char *TheFilter, void (*CBfunction)(char *),
		 DialogObject *ParentDialog)
{
	static int FirstTime = True;

	/* The core File Selector function : Create the form dialog and */
	/* initialise the internal selector variables 			*/

	if (TheFileSelectorDialog != NULL)
		FormFree(TheFileSelectorDialog); 

	TheFileSelectorDialog = FormCreate(10, 5, 245, 184, ParentDialog);

	/* First time in copy the default file path and filter */
	if (FirstTime) {
		FirstTime = !FirstTime;
		strcpy(CurrentPath,   ThePath);
		strcpy(CurrentFilter, TheFilter);
	}

	/* Initialize the input box contents */
	strcpy(PathBuffer, CurrentPath);
	strcpy(FilterBuffer, CurrentFilter);

	FormAddTextItem(TheFileSelectorDialog, 10, 24, "Directories");
	FormAddTextItem(TheFileSelectorDialog, 130, 24, "Files");

	/* Initialise the List boxes */

	if ((CurrentDirectories = FileSelectorGetDirList(CurrentPath)) != NULL) {
		DirListBox = FormAddListBoxItem(TheFileSelectorDialog, 10, 34, 
				   CurrentDirectories->NoFiles, 16, 12, 
				   CurrentDirectories->FileNames, CBFileSelectorSelectDir);
	}
	if ((CurrentFiles = FileSelectorGetFileList(CurrentPath, CurrentFilter)) != NULL) {
		FileListBox = FormAddListBoxItem(TheFileSelectorDialog, 130, 34, 
				   CurrentFiles->NoFiles, 16, 12, 
				   CurrentFiles->FileNames, CBFileSelectorSelectFile);
	}
		
	FormAddTextItem(TheFileSelectorDialog, 10, 10, "Filter");
	FilterInputBox = FormAddInputBoxItem(TheFileSelectorDialog, 48, 7, sizeof(FilterBuffer), 30, 
			FilterBuffer, CBFileSelectorSelectFilter);

	FormAddTextItem(TheFileSelectorDialog, 10, 146, "File");
	PathInputBox = FormAddInputBoxItem(TheFileSelectorDialog, 48, 143, 
				       sizeof(PathBuffer), 30, 
				       PathBuffer, CBFileSelectorSelectFilePath);

	FormAddButtonItem(TheFileSelectorDialog, 10, 164, "  OK  ", FORM_FUNCTION_BUTTON, CBFileSelectorOK);
/*	FormAddButtonItem(TheFileSelectorDialog, 100, 164, "Filter", FORM_FUNCTION_BUTTON, NULL);*/
	FormAddButtonItem(TheFileSelectorDialog, 188, 164, "Cancel", FORM_FUNCTION_BUTTON, CBFileSelectorCancel);


	SelectedFile[0] = '\0';
	CBuser = CBfunction;

	DialogSetActiveDialogObject(TheFileSelectorDialog);
}

void
FileSelectorClose(void)
{
	/* Free up the memory used by the file selector */

	FileSelectorFreeFileList(CurrentDirectories);
	FileSelectorFreeFileList(CurrentFiles);
	FormFree(TheFileSelectorDialog);
	TheFileSelectorDialog = NULL;
}

FileList *
FileSelectorGetDirList(char *ThePath) 
{
	/* Search ThePath and return all the directories  */

	DIR *TheDirPtr;
	struct dirent *TheDirEntry;
	struct stat    stbuff;	
	char FileBuffer[BUFSIZ];

	static FileList AFileList;

	int NoDirs = 0;

#ifdef DEBUG
	simcoupe_mesg(SIMCOUPE_INFO,"Open %s",ThePath);
#endif
	/* First Open the Directory */
	
	if ((TheDirPtr = opendir(ThePath)) != NULL) {
	
		/* Read all the file entries */
		while ((TheDirEntry = readdir(TheDirPtr)) != NULL && NoDirs < MAX_DIRECTORIES) {
			/* complete the path and test if the file is a directory */
#ifndef __DJGPP
			sprintf(FileBuffer,"%s/%s", ThePath, TheDirEntry->d_name);
#else
			sprintf(FileBuffer,"%s\\%s", ThePath, TheDirEntry->d_name);
#endif
			if (!stat(FileBuffer, &stbuff)) {
				if ((stbuff.st_mode & S_IFMT) == S_IFDIR &&
				    (stbuff.st_mode & S_IRUSR) && 
				    (strcmp(TheDirEntry->d_name, ".") != 0) && 
				   !(!strcmp(TheDirEntry->d_name, "..") &&
#ifndef __DJGPP
				     !strcmp(ThePath, "/")
#else
				     !strcmp(ThePath+1, ":\\"))
#endif
				   ) {
					/* Directory */
#ifdef DEBUG
					simcoupe_mesg(SIMCOUPE_INFO,"NoDirs = %d, Entry = %s",NoDirs,TheDirEntry->d_name);
#endif
					if (AFileList.FileNames[NoDirs] = (char *)malloc(1+strlen(TheDirEntry->d_name))) {
						strcpy(AFileList.FileNames[NoDirs++], TheDirEntry->d_name);
					}
#ifdef DEBUG
				simcoupe_mesg(SIMCOUPE_INFO,"!NoDirs = %d, Entry = %s",NoDirs,TheDirEntry->d_name);
#endif
				}
			}	
				
		}
		simcoupe_mesg(SIMCOUPE_INFO,"Close %s",ThePath);
		closedir(TheDirPtr);

		AFileList.NoFiles = NoDirs;
		return(&AFileList);
	}
	else if (!strcmp(ThePath, FDI_FLOPPY_1) || !strcmp(ThePath, FDI_FLOPPY_2)) { 
		/* Special pathname for real floppy drives */
		AFileList.FileNames[0] = (char *)malloc(1+strlen(NA));
		strcpy(AFileList.FileNames[0], NA);
		AFileList.NoFiles = 1;
		return(&AFileList);
	}
	else {
		return(NULL);
	}
}

FileList *
FileSelectorGetFileList(char *ThePath, char *TheFilter) 
{
	/* Search ThePath and return all the files which match TheFilter */

	DIR *TheDirPtr;
	struct dirent *TheDirEntry;
	struct stat    stbuff;	
	char FileBuffer[BUFSIZ];
	char *FilterList; 
	char *FilterPtr;
	static FileList AFileList;

	int NoFiles = 0;

	/* First Open the Directory */
	
	if ((TheDirPtr = opendir(ThePath)) != NULL) {
	
		/* Allocate memory for the Filter list */
		FilterList = (char *)malloc(1+strlen(TheFilter));

		/* Read all the file entries */
		while ((TheDirEntry = readdir(TheDirPtr)) != NULL && NoFiles < MAX_FILES) {
			/* complete the path and test if the file is a directory */
#ifndef __DJGPP
			sprintf(FileBuffer,"%s/%s", ThePath, TheDirEntry->d_name);
#else
			sprintf(FileBuffer,"%s\\%s", ThePath, TheDirEntry->d_name);
#endif
			if (!stat(FileBuffer, &stbuff)) {
				if ((stbuff.st_mode & S_IFMT) != S_IFDIR &&
				    (stbuff.st_mode & S_IRUSR)) {
					/* File */
					strcpy(FilterList, TheFilter);
					FilterPtr = strtok(FilterList, ",");
					while (FilterPtr) {
						if (!fnmatch(FilterPtr, FileBuffer, 0)) {
							/* Matched Filter Pattern */
							if (AFileList.FileNames[NoFiles] = (char *)malloc(1+strlen(TheDirEntry->d_name))) {
								strcpy(AFileList.FileNames[NoFiles++], TheDirEntry->d_name);
							}	
						}
						FilterPtr = strtok(NULL, ",");
					}
				}
			}	
				
		}
		closedir(TheDirPtr);
		free(FilterList);

		AFileList.NoFiles = NoFiles;
		return(&AFileList);
	}
	else if (!strcmp(ThePath, FDI_FLOPPY_1) || !strcmp(ThePath, FDI_FLOPPY_2)) { 
		/* Special pathname for real floppy drives */

		AFileList.FileNames[0] = (char *)malloc(1+strlen(NA));
		strcpy(AFileList.FileNames[0], NA);
		AFileList.NoFiles = 1;
		return(&AFileList);
	}
	else {
		return(NULL);
	}

}


void
FileSelectorFreeFileList(FileList *TheFileList)
			  
{
	/* Free up the storage allocated to the (Directory/File)List */
	int i;
	for (i=0; i < TheFileList->NoFiles; i++)
		free(TheFileList->FileNames[i]);
	TheFileList->NoFiles = 0;
}


/* FileSelector Internal Call Backs */

void
CBFileSelectorCancel(void)
{
	/* Call Back for the OK button : Close the Selector dialog and */
	/* exit after unmapping the dialog		       				   */

	UnMapDialog(TheFileSelectorDialog);
	DialogSetActiveDialogToParent(); 
	FileSelectorClose();

}

void
CBFileSelectorOK(void)
{
	/* Call Back for the Cancel button : Close the Selector dialog and */
	/* exit via the user Call back function CBuser		       */

	UnMapDialog(TheFileSelectorDialog);
	DialogSetActiveDialogToParent(); 
	FileSelectorClose();
	if (CBuser != NULL)
		CBuser(SelectedFile);
}



void
CBFileSelectorSelectDir(char *TheDir)
{
	/* Call Back for the Directory list box : change the working */
	/* directory and update the list boxes			     */

	if (!strncmp(TheDir, NA, 3))
		 return;

	if (strlen(CurrentPath)+strlen(TheDir) < MAX_PATH) {
		/* Prevent occurences of 'x/../y' in the path */
		if (strcmp(TheDir, "..") !=0 ) {
#ifndef __DJGPP
			/* The root path ("/") is a special case */
			if (strcmp(CurrentPath, "/") != 0)
				sprintf(CurrentPath, "%s/%s",CurrentPath, TheDir);
			else
				sprintf(CurrentPath, "/%s", TheDir);
#else
			/* The root path ("x:\") is a special case */
			if (strcmp(CurrentPath+1, ":\\") != 0)
				sprintf(CurrentPath, "%s\\%s",CurrentPath, TheDir);
			else
				sprintf(CurrentPath+1, ":\\%s", TheDir);
#endif
		}
		else {
#ifndef __DJGPP
			char *LastSeparator;
			if (strcmp(CurrentPath, "/") != 0) {
				LastSeparator = strrchr(CurrentPath, '/');
				if (LastSeparator != NULL) {
					if ((LastSeparator-CurrentPath) != 1)
						*(LastSeparator) = '\0';
					else
						/* Back to root, leave the '/' */
						*(LastSeparator+1) = '\0';
					
				}
			}
#else
			char *LastSeparator;
			if (strcmp(CurrentPath+1, ":\\") != 0) {
				LastSeparator = strrchr(CurrentPath, '\\');
				if (LastSeparator != NULL) {
					if ((LastSeparator-CurrentPath) != 2)
						*(LastSeparator) = '\0';
					else
						/* Back to root, leave the '\' */
						*(LastSeparator+1) = '\0';
					
				}			
			}

#endif
		}
		
		FileSelectorFreeFileList(CurrentDirectories);
		FileSelectorFreeFileList(CurrentFiles);
		if ((CurrentDirectories = FileSelectorGetDirList(CurrentPath)) != NULL &&
		    (CurrentFiles = FileSelectorGetFileList(CurrentPath, CurrentFilter)) != NULL) {
		}
		FormUpdateListBoxItem(TheFileSelectorDialog, DirListBox, 
					CurrentDirectories->NoFiles,
					CurrentDirectories->FileNames);
		FormUpdateListBoxItem(TheFileSelectorDialog, FileListBox,
					CurrentFiles->NoFiles,
					CurrentFiles->FileNames);
		strcpy(PathBuffer, CurrentPath);
		FormUpdateInputBoxItem(TheFileSelectorDialog, PathInputBox,
				       0, strlen(PathBuffer), PathBuffer);
	}	

}

void
CBFileSelectorSelectFile(char *TheFile)
{

	if (!strncmp(TheFile, NA, 3))
		 return;

	if (1+strlen(CurrentPath)+strlen(CurrentPath) < MAX_PATH) {

#ifndef __DJGPP
			/* The root path ("/") is a special case */
			if (strcmp(CurrentPath, "/") != 0)
				sprintf(SelectedFile, "%s/%s",CurrentPath, TheFile);
			else
				sprintf(SelectedFile, "/%s", TheFile);
#else
			/* The root path ("x:\") is a special case */
			if (strcmp(CurrentPath+1, ":\\") != 0)
				sprintf(SelectedFile, "%s\\%s",CurrentPath, TheFile);
			else
				sprintf(SelectedFile, "%s%s", CurrentPath, TheFile);
#endif

		FormUpdateListBoxItem(TheFileSelectorDialog, FileListBox,
					CurrentFiles->NoFiles,
					CurrentFiles->FileNames);
		strcpy(PathBuffer, SelectedFile);
		FormUpdateInputBoxItem(TheFileSelectorDialog, PathInputBox,
				       0, strlen(PathBuffer), PathBuffer);

	}

}


void
CBFileSelectorSelectFilter(char *TheFilter)
{
	strcpy(CurrentFilter, TheFilter);
	strcpy(FilterBuffer, CurrentFilter);
	FormUpdateInputBoxItem(TheFileSelectorDialog, FilterInputBox,
			       0, strlen(FilterBuffer), FilterBuffer);

	FileSelectorFreeFileList(CurrentDirectories);
	FileSelectorFreeFileList(CurrentFiles);
	if ((CurrentDirectories = FileSelectorGetDirList(CurrentPath)) != NULL &&
	    (CurrentFiles = FileSelectorGetFileList(CurrentPath, CurrentFilter)) != NULL) {
	}
	FormUpdateListBoxItem(TheFileSelectorDialog, DirListBox, 
						CurrentDirectories->NoFiles,
						CurrentDirectories->FileNames);
	FormUpdateListBoxItem(TheFileSelectorDialog, FileListBox,
						CurrentFiles->NoFiles,
						CurrentFiles->FileNames);

}

char *
FileSelectorCleanPath( char *ThePath )
{
	char *PathPtr;

	/* Cosmetic stuff : for bare drive ids (MSDOS only) add a   */
	/* directory delimiter (X: -> X:\). Remove extra delimiters */
	/* from the end of pathnames (/foo/bar/ -> /foo/bar). For   */
	/* MSDOS convert '/' to '\' to avoid confusion.             */

	
#ifdef __DJGPP	
	PathPtr = ThePath;
	while (*PathPtr) {
		if (*PathPtr == '/') {
			*PathPtr = '\\';
		}
		PathPtr++;	
	}
#endif	
	PathPtr = ThePath;
	while (*PathPtr) {
		/* Look for and remove occurences of '\\' or '//' */
#ifndef __DJGPP
		if (!strncmp(PathPtr, "//", 2)) {
#else
		if (!strncmp(PathPtr, "\\\\", 2)) {
#endif
			strcpy(PathPtr, PathPtr+1);
		}
		
		/* Insert a '\0' if the Path ends with a '/' or '\' */ 
#ifndef __DJGPP
		else if (!*(PathPtr+1) && (*PathPtr == '/')) {
#else
		else if (!*(PathPtr+1) && (*PathPtr == '\\')) {
#endif
			*PathPtr = '\0';
		}
		else {
			PathPtr++;
		}
	}	

#ifdef __DJGPP
	if ((strlen(ThePath) == 2) && (ThePath[1] == ':')) {
		strcat(ThePath, "\\");
	} 
#endif	

	return ThePath;
}

void
CBFileSelectorSelectFilePath(char *ThePath)
{
	/* Call Back function for Pathname input box */ 

	struct stat stbuff;	

	FileSelectorCleanPath(ThePath);

	/* The action depends on whether the path is a file or a directory */

	if (!stat(ThePath, &stbuff)) {
		if ((stbuff.st_mode & S_IFMT) == S_IFDIR &&
		    (stbuff.st_mode & S_IRUSR)  && strlen(ThePath)) {
			/* Directory */
			strcpy(CurrentPath, ThePath);

			FileSelectorFreeFileList(CurrentDirectories);
			FileSelectorFreeFileList(CurrentFiles);
			if ((CurrentDirectories = FileSelectorGetDirList(CurrentPath)) != NULL	&&
			    (CurrentFiles = FileSelectorGetFileList(CurrentPath, CurrentFilter)) != NULL) {
			}

			FormUpdateListBoxItem(TheFileSelectorDialog, DirListBox, 
						CurrentDirectories->NoFiles,
						CurrentDirectories->FileNames);
			FormUpdateListBoxItem(TheFileSelectorDialog, FileListBox,
						CurrentFiles->NoFiles,
						CurrentFiles->FileNames);
			strcpy(PathBuffer, CurrentPath);
			FormUpdateInputBoxItem(TheFileSelectorDialog, PathInputBox,
					       0, strlen(PathBuffer), PathBuffer);
		}
		else if ((stbuff.st_mode & S_IFMT) != S_IFDIR &&
			 (stbuff.st_mode & S_IRUSR) ) {
			/* File */	
			strcpy(SelectedFile, ThePath);
			CBFileSelectorOK();
		}
	}
	else if (!strcmp(ThePath, FDI_FLOPPY_1) || !strcmp(ThePath, FDI_FLOPPY_2)) { 
		/* Special pathname for real floppy drives */
		strcpy(SelectedFile, ThePath);
		CBFileSelectorOK();
	}
	else {
		/* Invalid input, just redraw the Pathname input box */	
		strcpy(PathBuffer, ThePath);
		FormUpdateInputBoxItem(TheFileSelectorDialog, PathInputBox,
				       0, strlen(PathBuffer), PathBuffer);
	}
}
