/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File form.c : User interface for SimCoupe : Form Object functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"

#include "dialog.h"

/* Form component data structures */

#define SCROLL_BAR_WIDTH	6

typedef struct _FormInputBox {

	int IsPressed;
	int TextWidth;
	uns8 Background;
	int ScrollOffset;
	int MaxNoChars;
	char *InputPtr;
	int PointPosition;
	void (*InputBoxCB)(char *InputText);

} FormInputBox;

typedef struct _FormListBox {

	int  IsPressed;
	int  TextHeight;
	int  TextWidth;
	uns8 Background;
	int  HasScrollBar;
	int  ScrollOffset;
	int  NoListItems;
	char **ListPtr;
	int  HighLightedItem;
	void (*ListBoxCB)(char *SelectedItem);

} FormListBox;

typedef struct _FormSwitch {

	int  IsPressed;
	int  OnValue;
	int  OffValue;
	int *ValuePtr;

} FormSwitch;

#define BUTTON_LABEL_LENGTH	50

typedef struct _FormButton {

	int IsPressed;
	char Label[BUTTON_LABEL_LENGTH];
	void *ActionPtr;

} FormButton;

/* Forward function prototypes */

void FormInputBoxDrawPoint(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos);
void FormListBoxHighLightItem(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos);

/* Form Object routines */

DialogObject *
FormCreate(uns8 x, uns8 y, uns8 width, uns8 height, DialogObject *TheParent)
{
	/* Create a Form. */
	
	DialogObject *DialogPtr;
	
	DialogPtr = CreateDialogObject(FORM_OBJECT, x, y, width, height,  TheParent);
	
	DialogPtr -> ObjectUnion.AForm.PassiveItemList = NULL;
	DialogPtr -> ObjectUnion.AForm.ActiveItemList = NULL;
	DialogPtr -> ObjectUnion.AForm.CurrentActiveItem = NULL;
	DialogPtr -> ObjectUnion.AForm.CurrentItemIsSelected = False;
	return(DialogPtr);
	
}

void
FormFree(DialogObject * TheDialog)
{
	/* Free up all the memory used by a form, and the Free the Object itself */
	
	FormObject *FormPtr = &(TheDialog->ObjectUnion.AForm);
	FormItem   *FormItemPtr;
	int Item;
	
	/* Free Passive Items */

	FormItemPtr = FormPtr->PassiveItemList;
	while (FormItemPtr != NULL) {
		FormItem *TmpPtr = FormItemPtr;

		FormItemPtr = FormItemPtr->NextItem; 
		free(TmpPtr->ItemPtr);
		free(TmpPtr);
	}

	/* Free Active Items */

	FormItemPtr = FormPtr->ActiveItemList; 
	if (FormItemPtr != NULL) {
		do {
			FormItem *TmpPtr = FormItemPtr;
	
			FormItemPtr = FormItemPtr->NextItem;
			free(TmpPtr->ItemPtr);
			free(TmpPtr);
		}
		while (FormItemPtr != FormPtr->ActiveItemList); 
	}
	FreeDialogObject(TheDialog);
	
}

FormItem*
FormAddTextItem(DialogObject *TheDialog, uns8 DeltaX, uns8 DeltaY, char *TheText)
{
	/* Add an text item to the form, adjusting the size appropriately */

	FormObject *FormPtr = &(TheDialog->ObjectUnion.AForm);
	FormItem   *FormItemPtr, *NewItemPtr;
	
	/* Create a new FormItem */
	NewItemPtr = (FormItem *)malloc(sizeof(FormItem));


	NewItemPtr->ItemType = FORM_TEXT;
	NewItemPtr->ItemPtr = (void *)malloc(1+strlen(TheText));
	strcpy(NewItemPtr->ItemPtr, TheText);

	NewItemPtr->DeltaX = DeltaX;
	NewItemPtr->DeltaY = DeltaY;
	NewItemPtr->height = GLYPH_HEIGHT;
	NewItemPtr->width  = GLYPH_WIDTH*strlen(TheText);
	NewItemPtr->NextItem = NULL;
	NewItemPtr->LastItem = NULL;

	/* Find the end of the Passive Item List and add the item */
	
	FormItemPtr = FormPtr->PassiveItemList;
	if (FormItemPtr !=NULL) {
		while (FormItemPtr->NextItem != NULL) 
			FormItemPtr = FormItemPtr->NextItem;

		FormItemPtr->NextItem = NewItemPtr;	
	}
	else {
		FormPtr->PassiveItemList = NewItemPtr;
	}

	return (NewItemPtr);
}

void
FormAddActiveItem(FormObject *TheForm, FormItem *TheNewItem)
{
	/* Find the end of the Active Item List and add the new item */
	
	FormItem *FormItemPtr = TheForm->ActiveItemList;

	if (FormItemPtr !=NULL) {

		/* The Active list is cyclic, so we get to the end by */
		/* 'going back' from the start!			      */

		FormItem *EndItemPtr = TheForm->ActiveItemList->LastItem;		

		/* The item to be inserted points up to the (old) EndItem */
		/* and down to the top of the list */

		TheNewItem->LastItem = EndItemPtr;
		TheNewItem->NextItem = TheForm->ActiveItemList;
		
		/* Now insert the new item below the (old) EndItem */
		
		EndItemPtr->NextItem = TheNewItem;
		
		/* Finally complete the cyclic list by pointing the LastItem */
		/* on the list head to the new item			     */
		
		TheForm->ActiveItemList->LastItem = TheNewItem;
	}
	else {
		/* The first element points back to itself */
		
		TheNewItem->NextItem = TheNewItem;
		TheNewItem->LastItem = TheNewItem; 
	
		TheForm->ActiveItemList = TheNewItem;
		

	}
}

FormItem *
FormAddSwitchItem(DialogObject *TheDialog, uns8 DeltaX, uns8 DeltaY, 
				  uns8 Height, uns8 Width, 
				  int *ValuePtr, int OnValue, int OffValue)
{
	/* Add an Switch item to the form */

	FormObject *FormPtr = &(TheDialog->ObjectUnion.AForm);
	FormItem   *FormItemPtr, *NewItemPtr;
	FormSwitch *SwitchPtr;

	/* Create a Switch structure */

	SwitchPtr = (FormSwitch *)malloc(sizeof(FormSwitch));

	SwitchPtr->IsPressed = False;
	SwitchPtr->OnValue   = OnValue;
	SwitchPtr->OffValue  = OffValue;
	SwitchPtr->ValuePtr  = ValuePtr;

	if (SwitchPtr->ValuePtr) {

		/* Determine the initial state of the switch */
		if (*SwitchPtr->ValuePtr == OnValue) {
			SwitchPtr->IsPressed = True;
		}
		else {
			SwitchPtr->IsPressed = False;
		}
	}
	
	/* Create a new FormItem, linking in the Switch */
	NewItemPtr = (FormItem *)malloc(sizeof(FormItem));

	NewItemPtr->ItemType = FORM_SWITCH;
	NewItemPtr->ItemPtr = SwitchPtr;

	NewItemPtr->DeltaX = DeltaX;
	NewItemPtr->DeltaY = DeltaY;
	NewItemPtr->height = Height+(2*SH_WID);
	NewItemPtr->width  = Width +(2*SH_WID);
	NewItemPtr->NextItem = NULL;
	NewItemPtr->LastItem = NULL;

	FormAddActiveItem(FormPtr, NewItemPtr);

	return (NewItemPtr);

}

FormItem *
FormAddButtonItem(DialogObject *TheDialog, uns8 DeltaX, uns8 DeltaY, char *TheText, int ButtonType, void* ActionPtr)
{
	/* Add an button item to the form, adjusting the size appropriately */

	FormObject *FormPtr = &(TheDialog->ObjectUnion.AForm);
	FormItem   *FormItemPtr, *NewItemPtr;
	FormButton *ButtonPtr;

	/* Create a Button structure */

	ButtonPtr = (FormButton *)malloc(sizeof(FormButton));

	ButtonPtr->IsPressed = False;
	ButtonPtr->ActionPtr = ActionPtr;
	strcpy(ButtonPtr->Label, TheText);

	/* Create a new FormItem, linking in the Button */
	NewItemPtr = (FormItem *)malloc(sizeof(FormItem));


	NewItemPtr->ItemType = ButtonType;
	NewItemPtr->ItemPtr = ButtonPtr;

	NewItemPtr->DeltaX = DeltaX;
	NewItemPtr->DeltaY = DeltaY;
	NewItemPtr->height = GLYPH_HEIGHT + 2*(SH_WID + 1);
	NewItemPtr->width  = GLYPH_WIDTH*strlen(TheText) + 2*(SH_WID + 2);
	NewItemPtr->NextItem = NULL;
	NewItemPtr->LastItem = NULL;

	FormAddActiveItem(FormPtr, NewItemPtr);

	return (NewItemPtr);

}

FormItem *
FormAddListBoxItem(DialogObject *TheDialog, uns8 DeltaX, uns8 DeltaY, int NoItems, 
		   uns8 TextWidth, uns8 TextHeight, 
		   char **TheList, void (*CBFunction)(char *))
{
	/* Add an list box item to the form */

	FormObject  *FormPtr = &(TheDialog->ObjectUnion.AForm);
	FormItem    *FormItemPtr, *NewItemPtr;
	FormListBox *ListBoxPtr;

	/* Create a list box structure */

	ListBoxPtr = (FormListBox *)malloc(sizeof(FormListBox));

	ListBoxPtr->IsPressed  = False;
	ListBoxPtr->TextWidth  = TextWidth;
	ListBoxPtr->TextHeight = TextHeight;
	ListBoxPtr->HasScrollBar = (NoItems > TextHeight);
	ListBoxPtr->ScrollOffset = 0;
	ListBoxPtr->ListBoxCB = CBFunction;
	ListBoxPtr->ListPtr   = TheList;
	ListBoxPtr->HighLightedItem = 0;
	ListBoxPtr->NoListItems = NoItems;
	ListBoxPtr->Background = SHCOL_MEDIUM;

	/* Create a new FormItem, linking in the Button */
	NewItemPtr = (FormItem *)malloc(sizeof(FormItem));


	NewItemPtr->ItemType = FORM_LIST_BOX;
	NewItemPtr->ItemPtr = ListBoxPtr;

	NewItemPtr->DeltaX = DeltaX;
	NewItemPtr->DeltaY = DeltaY;
	NewItemPtr->height = GLYPH_HEIGHT*TextHeight + 2*(SH_WID + 1);
	NewItemPtr->width  = GLYPH_WIDTH *TextWidth  + 2*(SH_WID + 2);
	NewItemPtr->NextItem = NULL;
	NewItemPtr->LastItem = NULL;

	FormAddActiveItem(FormPtr, NewItemPtr);

	return (NewItemPtr);

}

void
FormUpdateListBoxItem(DialogObject *TheDialog, FormItem *TheFormItem, 
		      int NoItems, char **TheList)
{
	/* Function to allow the data in the List Box to be updated, using */
	/* the pointer to the item returned by FormAddListBoxItem()        */
	
	FormListBox *ListBoxPtr = TheFormItem->ItemPtr;
	
	ListBoxPtr->HasScrollBar = (NoItems > ListBoxPtr->TextHeight);
	ListBoxPtr->ScrollOffset = 0;
	ListBoxPtr->ListPtr   = TheList;
	ListBoxPtr->NoListItems = NoItems;

	if (ListBoxPtr->HighLightedItem > NoItems-1)
		ListBoxPtr->HighLightedItem = 0;

	FormDrawListBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);	
	if (ListBoxPtr->IsPressed)
		FormListBoxHighLightItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
}

FormItem *
FormAddInputBoxItem(DialogObject *TheDialog, uns8 DeltaX, uns8 DeltaY, int NoChars, 
		   uns8 TextWidth, 
		   char *InputDefault, void (*CBFunction)(char *))
{
	/* Add an input box item to the form */

	FormObject   *FormPtr = &(TheDialog->ObjectUnion.AForm);
	FormItem     *FormItemPtr, *NewItemPtr;
	FormInputBox *InputBoxPtr;

	/* Create a input box structure */

	InputBoxPtr = (FormInputBox *)malloc(sizeof(FormInputBox));

	InputBoxPtr->IsPressed  = False;
	InputBoxPtr->TextWidth  = TextWidth;
	InputBoxPtr->ScrollOffset = 0;
	InputBoxPtr->InputBoxCB = CBFunction;
	InputBoxPtr->InputPtr   = InputDefault;
	InputBoxPtr->PointPosition = 0;
	InputBoxPtr->MaxNoChars = NoChars;
	InputBoxPtr->Background = SHCOL_MEDIUM;

	/* Create a new FormItem, linking in the Input Box */
	NewItemPtr = (FormItem *)malloc(sizeof(FormItem));

	NewItemPtr->ItemType = FORM_INPUT_BOX;
	NewItemPtr->ItemPtr = InputBoxPtr;

	NewItemPtr->DeltaX = DeltaX;
	NewItemPtr->DeltaY = DeltaY;
	NewItemPtr->height = GLYPH_HEIGHT + 2*(SH_WID + 1);
	NewItemPtr->width  = GLYPH_WIDTH*TextWidth  + 2*(SH_WID + 2);
	NewItemPtr->NextItem = NULL;
	NewItemPtr->LastItem = NULL;

	FormAddActiveItem(FormPtr, NewItemPtr);

	return (NewItemPtr);

}

void
FormUpdateInputBoxItem(DialogObject *TheDialog, FormItem *TheFormItem, 
		      int ScrollOffset, int NoChars, char *InputDefault)
{
	/* Function to allow the data in the Input Box to be updated, using */
	/* the pointer to the item returned by FormAddInputBoxItem()        */
	
	FormInputBox *InputBoxPtr = TheFormItem->ItemPtr;
	
	if (ScrollOffset >= 0) 
		InputBoxPtr->ScrollOffset = ScrollOffset;
	else if ((strlen(InputDefault)-InputBoxPtr->TextWidth) > 0)
		InputBoxPtr->ScrollOffset = strlen(InputDefault)-InputBoxPtr->TextWidth;
	else
		InputBoxPtr->ScrollOffset = 0;

	InputBoxPtr->InputPtr   = InputDefault;
	InputBoxPtr->PointPosition = InputBoxPtr->ScrollOffset;
	InputBoxPtr->MaxNoChars = NoChars;

	FormDrawInputBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);	
	if (InputBoxPtr->IsPressed)
		FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);
}


void
FormDrawTextItem(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos)
{

	DrawString(FormXPos+TheFormItem->DeltaX, FormYPos+TheFormItem->DeltaY,
		   (char *)(TheFormItem->ItemPtr), SHCOL_BLACK, DM_OVER);

}

void
FormDrawSwitchItem(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos)
{
	/* Draw a Switch Item, reflecting the current state of the switch */

	FormSwitch *SwitchPtr = (FormSwitch *)TheFormItem->ItemPtr;
	
	if (SwitchPtr->IsPressed) {
		DrawShadowBox(FormXPos+TheFormItem->DeltaX, FormYPos+TheFormItem->DeltaY,
		      TheFormItem->width, TheFormItem->height, SHADOW_IN, True, 
		      DM_OVER);
	}
	else {
		DrawShadowBox(FormXPos+TheFormItem->DeltaX, FormYPos+TheFormItem->DeltaY,
		      TheFormItem->width, TheFormItem->height, SHADOW_OUT, True, 
		      DM_OVER);
	}
}

void
FormDrawButtonItem(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos)
{

	FormButton *ButtonPtr = (FormButton *)TheFormItem->ItemPtr;
	
	DrawShadowBox(FormXPos+TheFormItem->DeltaX, FormYPos+TheFormItem->DeltaY,
		      TheFormItem->width, TheFormItem->height, SHADOW_OUT, True, 
		      DM_OVER);

	DrawString(2*SH_WID+FormXPos+TheFormItem->DeltaX, 
		   2*SH_WID+FormYPos+TheFormItem->DeltaY,ButtonPtr->Label,
		   SHCOL_BLACK, DM_OVER);
}

void
FormDrawListBoxItem(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos)
{
	FormListBox *ListBoxPtr = (FormListBox *)TheFormItem->ItemPtr;
	int ListBoxWidth = TheFormItem->width;
	int ListBoxTextWidth = ListBoxPtr->TextWidth;
	char TextBuffer[256];
	int i;

	/* Adjust main list box for a scroll bar if present */
	if (ListBoxPtr->HasScrollBar) {
		ListBoxWidth -= (SCROLL_BAR_WIDTH + 6);
		ListBoxTextWidth -= 2;
	}
	
	DrawShadowBox(FormXPos+TheFormItem->DeltaX, FormYPos+TheFormItem->DeltaY,
		      ListBoxWidth, TheFormItem->height, SHADOW_IN, False, 
		      DM_OVER);
	FillArea(2+FormXPos+TheFormItem->DeltaX,
		 2+FormYPos+TheFormItem->DeltaY,
		 ListBoxWidth-4, TheFormItem->height-4, 
		 ListBoxPtr->Background, DM_OVER);



	/* Draw the list text, truncating each item if required */
	
	for (i = 0; i < ListBoxPtr->TextHeight; i++) {
	
		/* Just in case  TextHeight > NoListItems */
		if (i < ListBoxPtr->NoListItems) {
			strncpy(TextBuffer, ListBoxPtr->ListPtr[i+ListBoxPtr->ScrollOffset], 
				(size_t)(ListBoxTextWidth));
			TextBuffer[ListBoxTextWidth] = '\0';
			DrawString(2*SH_WID+FormXPos+TheFormItem->DeltaX, 
			   	   (GLYPH_HEIGHT*i)+2*SH_WID+FormYPos+TheFormItem->DeltaY,
			   	   TextBuffer, SHCOL_BLACK, DM_OVER);
		}
	}

	/* Draw the scroll bar in the appropriate position if required */

	if (ListBoxPtr->HasScrollBar) {
	
		int ScrollBarHeight = ((-4+TheFormItem->height) * ListBoxPtr->TextHeight) /
				  ListBoxPtr->NoListItems;
		int ScrollBarOffset = ((-4+TheFormItem->height) *  ListBoxPtr->ScrollOffset) /
				  ListBoxPtr->NoListItems;
		DrawShadowBox(FormXPos+TheFormItem->DeltaX+ListBoxWidth+2,
			      FormYPos+TheFormItem->DeltaY,
		      	      SCROLL_BAR_WIDTH+4, 
		      	      TheFormItem->height, SHADOW_IN, True, 
			      DM_OVER);
		DrawShadowBox(FormXPos+TheFormItem->DeltaX+ListBoxWidth+4,
			      FormYPos+TheFormItem->DeltaY+2+ScrollBarOffset,
			      SCROLL_BAR_WIDTH,
			      ScrollBarHeight, SHADOW_OUT, True,
			      DM_OVER);
	}
}

void
FormDrawInputBoxItem(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos)
{
	FormInputBox *InputBoxPtr = (FormInputBox *)TheFormItem->ItemPtr;
	char TextBuffer[256];

	DrawShadowBox(FormXPos+TheFormItem->DeltaX, FormYPos+TheFormItem->DeltaY,
		      TheFormItem->width, TheFormItem->height, SHADOW_IN, False, 
		      DM_OVER);
	FillArea(2+FormXPos+TheFormItem->DeltaX,
		 2+FormYPos+TheFormItem->DeltaY,
		 TheFormItem->width-4, TheFormItem->height-4, 
		 InputBoxPtr->Background, DM_OVER);


	/* Draw the input text, starting from the position ScrollOffset */
	
	strncpy(TextBuffer, InputBoxPtr->InputPtr+InputBoxPtr->ScrollOffset, InputBoxPtr->TextWidth);
	TextBuffer[InputBoxPtr->TextWidth] = '\0';
	DrawString(2*SH_WID+FormXPos+TheFormItem->DeltaX, 
		   2*SH_WID+FormYPos+TheFormItem->DeltaY,
		   TextBuffer, SHCOL_BLACK, DM_OVER);

}

void
FormDrawItem(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos)
{
	switch (TheFormItem->ItemType) {
		case FORM_TEXT:
			FormDrawTextItem(TheFormItem, FormXPos, FormYPos);
			break;
		case FORM_DIALOG_BUTTON: case FORM_FUNCTION_BUTTON:
			FormDrawButtonItem(TheFormItem, FormXPos, FormYPos);
			break;
		case FORM_SWITCH:
			FormDrawSwitchItem(TheFormItem, FormXPos, FormYPos);
			break;
		case FORM_LIST_BOX:
			FormDrawListBoxItem(TheFormItem, FormXPos, FormYPos);
			break;
		case FORM_INPUT_BOX:
			FormDrawInputBoxItem(TheFormItem, FormXPos, FormYPos);
			break;


	}

}

void
FormDraw(DialogObject *TheDialog)
{
	uns8 ItemX, ItemY;
	int  Item;

	FormObject *FormPtr = &(TheDialog->ObjectUnion.AForm);
	FormItem   *FormItemPtr;

	/* Draw the Form using the appropriate primitives   	      */
	/* The Object should be mapped before this function is called */

	/* First we draw a shadow box */
	
	DrawShadowBox(TheDialog->xpos,  TheDialog->ypos,
		      TheDialog->width, TheDialog->height,
		      SHADOW_OUT, True, DM_OVER);

	/* Now write in the Form entries */

	/* Passive Items */
	
	FormItemPtr = FormPtr->PassiveItemList;
	while (FormItemPtr != NULL) {
		FormDrawItem(FormItemPtr, TheDialog->xpos, TheDialog->ypos);
		FormItemPtr = FormItemPtr->NextItem;	
	}

	/* Active Items */
	
	FormItemPtr = FormPtr->ActiveItemList; 
	if (FormItemPtr != NULL) {
		do {
			FormDrawItem(FormItemPtr, TheDialog->xpos, TheDialog->ypos);
			FormItemPtr = FormItemPtr->NextItem;
		}
		while (FormItemPtr != FormPtr->ActiveItemList); 
	}
}

void
FormHighLight(DialogObject *TheDialog, FormItem *TheFormItem, int DoHighLight)
{
	/* Highlight or unhighlight a given Form Item */
	
	uns8 ItemX = (TheDialog->xpos+ TheFormItem->DeltaX) - SH_WID;
	uns8 ItemY = (TheDialog->ypos+ TheFormItem->DeltaY) - SH_WID;

	if (!DoHighLight)
		DrawBox(ItemX, ItemY, 
			TheFormItem->width  + 2*SH_WID, 
			TheFormItem->height + 2*SH_WID,
			SHCOL_MEDIUM, DM_OVER);
	else
		DrawBox(ItemX, ItemY, 
			TheFormItem->width  + 2*SH_WID, 
			TheFormItem->height + 2*SH_WID,
			SHCOL_DARK, DM_OVER);
 
}


int
FormButtonCore(DialogObject *TheDialog, FormItem *TheFormItem)
{
	/* Deal with Button events */

	FormButton *ButtonPtr = (FormButton *)TheFormItem->ItemPtr;
	void (*ItemFunction)(void);

	if (!ButtonPtr->IsPressed) {
		/* First time : Push button down and do the action */
		ButtonPtr->IsPressed = True;
		DrawShadowBox(TheDialog->xpos+TheFormItem->DeltaX, 
			      TheDialog->ypos+TheFormItem->DeltaY,
			      TheFormItem->width, TheFormItem->height, SHADOW_IN, False, 
			      DM_OVER);

		if (ButtonPtr->ActionPtr != NULL) {
		
			switch (TheFormItem->ItemType) {
			
			case FORM_DIALOG_BUTTON:		
				ActiveDialogObject = (DialogObject *)ButtonPtr->ActionPtr;
				break;
			case FORM_FUNCTION_BUTTON:		
				ItemFunction = (void (*)(void))ButtonPtr->ActionPtr;
				ItemFunction();
				break;
			}
		}
		return True;
	}
	else {
		/* Last time : Pop button back and signal unselected */
		ButtonPtr->IsPressed = False;
		DrawShadowBox(TheDialog->xpos+TheFormItem->DeltaX, 
			      TheDialog->ypos+TheFormItem->DeltaY,
			      TheFormItem->width, TheFormItem->height, SHADOW_OUT, False, 
			      DM_OVER);
		return False;

	}
}

int
FormSwitchCore(DialogObject *TheDialog, FormItem *TheFormItem)
{
	/* Deal with Switch events */

	FormSwitch *SwitchPtr = (FormSwitch *)TheFormItem->ItemPtr;
	void (*ItemFunction)(void);

	if (!SwitchPtr->IsPressed) {
		/* Switch On : Push button down and Set *ValuePtr to OnValue */
		SwitchPtr->IsPressed = True;
		DrawShadowBox(TheDialog->xpos+TheFormItem->DeltaX, 
			      TheDialog->ypos+TheFormItem->DeltaY,
			      TheFormItem->width, TheFormItem->height, SHADOW_IN, False, 
			      DM_OVER);

		if (SwitchPtr->ValuePtr != NULL) {
			*SwitchPtr->ValuePtr = SwitchPtr->OnValue;
		}
	}
	else {
		/* Switch Off : Pop button back and Set *ValuePtr to OffValue */
		SwitchPtr->IsPressed = False;
		DrawShadowBox(TheDialog->xpos+TheFormItem->DeltaX, 
			      TheDialog->ypos+TheFormItem->DeltaY,
			      TheFormItem->width, TheFormItem->height, SHADOW_OUT, False, 
			      DM_OVER);

		if (SwitchPtr->ValuePtr != NULL) {
			*SwitchPtr->ValuePtr = SwitchPtr->OffValue;
		}
	}
	/* Switches are never left active - they */
	/* are always reactivated by the user */
	return False;
}

void
FormListBoxHighLightItem(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos)
{
	/* Highlight/UnHightlight the required item in a List Box */

	FormListBox *ListBoxPtr = (FormListBox *)TheFormItem->ItemPtr;
	int HighLightOffset = ListBoxPtr->HighLightedItem - ListBoxPtr->ScrollOffset;
	int ListBoxWidth = TheFormItem->width;

	/* Adjust main list box for a scroll bar if present */
	if (ListBoxPtr->HasScrollBar)
		ListBoxWidth -= (SCROLL_BAR_WIDTH + 6);

	if (HighLightOffset>=0 && HighLightOffset < ListBoxPtr->TextHeight && ListBoxPtr->NoListItems) {
	
		FillArea(2*SH_WID+FormXPos+TheFormItem->DeltaX,
			 2*SH_WID+FormYPos+TheFormItem->DeltaY+(HighLightOffset*GLYPH_HEIGHT),
			 ListBoxWidth-4*SH_WID, GLYPH_HEIGHT,
			 0x0f, DM_XOR);			 
	
	
	}
}

int
FormListBoxCore(DialogObject *TheDialog, FormItem *TheFormItem)
{
	/* Deal with List Box events */

	FormListBox *ListBoxPtr = (FormListBox *)TheFormItem->ItemPtr;
	int ListBoxActive;
	int KeyEvent;
	int TheButtonState;
	int TheTabState;

	ListBoxActive = True;

	if (!ListBoxPtr->IsPressed) {
		/* First time : Change the list background and redraw the list box */
		ListBoxPtr->IsPressed  = True;
		ListBoxPtr->Background = SHCOL_WHITE;
		FormDrawListBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
		FormListBoxHighLightItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
	}

	/* Deal with Events */
	
	/* Keyboard events */

	switch (KeyEvent = GetSpecialKey()) {
	
		case KEY_UP:
			if (ListBoxPtr->HighLightedItem) {
				FormListBoxHighLightItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
				if (ListBoxPtr->HasScrollBar && 
				   !(ListBoxPtr->HighLightedItem - ListBoxPtr->ScrollOffset)) {
					/* Scroll Up */
					(ListBoxPtr->ScrollOffset)--;
					FormDrawListBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
				}
				(ListBoxPtr->HighLightedItem)--;
				FormListBoxHighLightItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);				

			}
			break;

		case KEY_DOWN:
			if (ListBoxPtr->HighLightedItem+1 < ListBoxPtr->NoListItems) {
				FormListBoxHighLightItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);			
				if (ListBoxPtr->HasScrollBar && 
				    (1+ListBoxPtr->HighLightedItem - ListBoxPtr->ScrollOffset == ListBoxPtr->TextHeight)) {
					/* Scroll Down */			
					(ListBoxPtr->ScrollOffset)++;
					FormDrawListBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);

				}
				(ListBoxPtr->HighLightedItem)++;
				FormListBoxHighLightItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
			}
			break;
		case KEY_ENTER:
			if (ListBoxPtr->ListBoxCB != NULL) {
				FormListBoxHighLightItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
				ListBoxPtr->ListBoxCB(ListBoxPtr->ListPtr[ListBoxPtr->HighLightedItem]);
			}
			break;
		case KEY_ESCAPE: case KEY_LEFT: case KEY_RIGHT: 
			ListBoxActive = False;	
			break;					
	
	}
	if (GetTabKeyState(&TheTabState)) {
		ListBoxActive = TheTabState;
	}
	else  {
	/* Mouse events */

	  if (PointerGetButtonState(&TheButtonState) & BUTTON_1) {
		ListBoxActive = !(!(TheButtonState & BUTTON_1) &&
				       !PointerInsideRegion( 
					(TheDialog->xpos+ TheFormItem->DeltaX) - SH_WID,
					(TheDialog->ypos+ TheFormItem->DeltaY) - SH_WID,
					TheFormItem->width  + 2*SH_WID, 
					TheFormItem->height + 2*SH_WID));
	  }
	}	

	/* Check for exit event */
	if (!ListBoxActive) {
		ListBoxPtr->IsPressed  = False;
		ListBoxPtr->Background = SHCOL_MEDIUM;
		FormDrawListBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
	}
	
	return (ListBoxActive);

}

void
FormInputBoxDrawPoint(FormItem *TheFormItem, uns8 FormXPos, uns8 FormYPos)
{
	/* Draw/Erase the point in a Input Box */

	FormInputBox *InputBoxPtr = (FormInputBox *)TheFormItem->ItemPtr;
	int PointOffset = InputBoxPtr->PointPosition - InputBoxPtr->ScrollOffset;

	if (PointOffset>=0 && PointOffset < InputBoxPtr->TextWidth && strlen(InputBoxPtr->InputPtr)) {
	
		DrawLineVert(2*SH_WID+FormXPos+TheFormItem->DeltaX+(GLYPH_WIDTH*PointOffset),
			     2*SH_WID+FormYPos+TheFormItem->DeltaY,
			     GLYPH_HEIGHT, 0x0f, DM_XOR);
	}
}


void
FormInputBoxPointRight(DialogObject *TheDialog, FormItem *TheFormItem)
{

	/* Move the Point one character to the right, dealing with scrolling if required */

	FormInputBox *InputBoxPtr = (FormInputBox *)TheFormItem->ItemPtr;

	if (InputBoxPtr->PointPosition < strlen(InputBoxPtr->InputPtr)) {
		FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);
		if ((1+InputBoxPtr->PointPosition - InputBoxPtr->ScrollOffset) == InputBoxPtr->TextWidth) {
			/* Scroll Right */			
			(InputBoxPtr->ScrollOffset)++;
			FormDrawInputBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);

		}
		(InputBoxPtr->PointPosition)++;
		FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);
	}
}

void
FormInputBoxPointLeft(DialogObject *TheDialog, FormItem *TheFormItem)
{
	/* Move the Point one character to the left, dealing with scrolling if required */

	FormInputBox *InputBoxPtr = (FormInputBox *)TheFormItem->ItemPtr;

	if (InputBoxPtr->PointPosition) {
		FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);
		if (!(InputBoxPtr->PointPosition - InputBoxPtr->ScrollOffset)) {
			/* Scroll Left */
			(InputBoxPtr->ScrollOffset)--;
			FormDrawInputBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
		}
		(InputBoxPtr->PointPosition)--;
		FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);				
	}

}
int
FormInputBoxCore(DialogObject *TheDialog, FormItem *TheFormItem)
{
	/* Deal with Input Box events */

	FormInputBox *InputBoxPtr = (FormInputBox *)TheFormItem->ItemPtr;
	int InputBoxActive;
	int KeyEvent;
	int TheButtonState;
	int TheTabState;
	char AsciiInput, InputBuffer[255];

	InputBoxActive = True;

	if (!InputBoxPtr->IsPressed) {
		/* First time : Change the box background and redraw the input box */
		InputBoxPtr->IsPressed  = True;
		InputBoxPtr->Background = SHCOL_WHITE;
		FormDrawInputBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
		FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);
	}

	/* Keyboard Events */

	switch (KeyEvent = GetSpecialKey()) {
	
		case KEY_LEFT:
			FormInputBoxPointLeft(TheDialog, TheFormItem);
			break;

		case KEY_RIGHT:
			FormInputBoxPointRight(TheDialog, TheFormItem);			break;
			break;
			
		case KEY_ENTER:
			if (InputBoxPtr->InputBoxCB != NULL) {
				FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);
				InputBoxPtr->InputBoxCB(InputBoxPtr->InputPtr);
			}
			break;		
		case KEY_ESCAPE: case KEY_UP:
			InputBoxActive = False;	
			break;

		case KEY_DELETE:
			if (InputBoxPtr->PointPosition && strlen(InputBoxPtr->InputPtr)) {		
				FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);

				strncpy(InputBuffer, InputBoxPtr->InputPtr,
					InputBoxPtr->PointPosition-1);
				strcpy(InputBuffer+InputBoxPtr->PointPosition-1,
					InputBoxPtr->InputPtr+InputBoxPtr->PointPosition);
				strcpy(InputBoxPtr->InputPtr, InputBuffer);
				FormDrawInputBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
				FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);

				FormInputBoxPointLeft(TheDialog, TheFormItem);
			
			}
			else if (!InputBoxPtr->PointPosition) {
				/* Delete on at Lhs -> clear input box */
				FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);

				*(InputBoxPtr->InputPtr) = '\0';
				
				FormDrawInputBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
				FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);
			}
			break;
	
		default:
			if ((AsciiInput = GetAsciiKey()) && 
			    InputBoxPtr->PointPosition <= strlen(InputBoxPtr->InputPtr)) {

				FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);

				strncpy(InputBuffer, InputBoxPtr->InputPtr,
					InputBoxPtr->PointPosition);
				InputBuffer[InputBoxPtr->PointPosition] = AsciiInput;
				strcpy(InputBuffer+1+InputBoxPtr->PointPosition,
					InputBoxPtr->InputPtr+InputBoxPtr->PointPosition);
				strcpy(InputBoxPtr->InputPtr, InputBuffer);
				FormDrawInputBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
				FormInputBoxDrawPoint(TheFormItem, TheDialog->xpos, TheDialog->ypos);

				FormInputBoxPointRight(TheDialog, TheFormItem);
			}			

			break;

	}
	if (GetTabKeyState(&TheTabState)) {
		InputBoxActive = TheTabState;
	}
	else { 
	/* Mouse events */
	  if (PointerGetButtonState(&TheButtonState) & BUTTON_1) {
		InputBoxActive = !(!(TheButtonState & BUTTON_1) &&
				       !PointerInsideRegion( 
					(TheDialog->xpos+ TheFormItem->DeltaX) - SH_WID,
					(TheDialog->ypos+ TheFormItem->DeltaY) - SH_WID,
					TheFormItem->width  + 2*SH_WID, 
					TheFormItem->height + 2*SH_WID));
	  }	
	}

	/* Check for exit event */
	if (!InputBoxActive) {
		InputBoxPtr->IsPressed  = False;
		InputBoxPtr->Background = SHCOL_MEDIUM;
		FormDrawInputBoxItem(TheFormItem, TheDialog->xpos, TheDialog->ypos);
	}
	
	return (InputBoxActive);

}


int 
FormItemCore(DialogObject *TheDialog, FormItem *TheFormItem)
{
	/* Deal with Form Item events. Call the appropriate function */

	int ItemStillSelected = False;

	switch (TheFormItem->ItemType) {

		case FORM_DIALOG_BUTTON: case FORM_FUNCTION_BUTTON:
			ItemStillSelected = FormButtonCore(TheDialog, TheFormItem);
			break;
		case FORM_SWITCH:
			ItemStillSelected = FormSwitchCore(TheDialog, TheFormItem);
			break;
		case FORM_LIST_BOX: 
			ItemStillSelected = FormListBoxCore(TheDialog, TheFormItem);
			break;
		case FORM_INPUT_BOX: 
			ItemStillSelected = FormInputBoxCore(TheDialog, TheFormItem);
			break;
	}

	return ItemStillSelected;
}

void
FormCore(DialogObject *TheDialog)
{
	FormObject *FormPtr = &(TheDialog->ObjectUnion.AForm);
	int KeyEvent = KEY_NONE;

	uns8 ItemX, ItemY;
	int Item, FormFoundOption;
	int TheButtonState;
	int TheTabState;
	FormItem* FormItemPtr;

	/* The core form code, called by DialogCore */
	
	/* Find the state of the TAB key once per cycle */
	SetTabKeyState();

	/* Map and draw the form if required */
	
	if (!TheDialog -> IsMapped) {
		MapDialog(TheDialog);
		FormDraw(TheDialog);
		if (FormPtr->ActiveItemList != NULL) {
			FormPtr->CurrentActiveItem=FormPtr->ActiveItemList;
			FormHighLight(TheDialog, FormPtr->CurrentActiveItem, True);
		}
	}

	/* Deal with Events */

	if (FormPtr->ActiveItemList != NULL) {

		/* Deal with Active Items on the form */	
	
		if (FormPtr->CurrentItemIsSelected) {
			/* We must check whether the Dialog Object is still valid, just */
			/* in case it is freed somewhere in the FormItemCore() tree     */
	
			int IsSelected = FormItemCore(TheDialog, FormPtr->CurrentActiveItem);
			if (TheDialog != NULL)
				FormPtr->CurrentItemIsSelected = IsSelected;
			else
				return;
		}
		
		if (!FormPtr->CurrentItemIsSelected) {

			/* Keyboard Events */
			KeyEvent=GetSpecialKey();
			switch (KeyEvent) {
		
			case KEY_LEFT:
				FormHighLight(TheDialog, FormPtr->CurrentActiveItem, False);
				FormPtr->CurrentActiveItem = FormPtr->CurrentActiveItem->LastItem;	
				FormHighLight(TheDialog, FormPtr->CurrentActiveItem, True);
				break;
			case KEY_RIGHT:
				FormHighLight(TheDialog, FormPtr->CurrentActiveItem, False);
				FormPtr->CurrentActiveItem = FormPtr->CurrentActiveItem->NextItem;	
				FormHighLight(TheDialog, FormPtr->CurrentActiveItem, True);
				break;
			case KEY_ENTER: case KEY_DOWN:
				FormPtr->CurrentItemIsSelected = True;
				break;
			}
		}
		if (GetTabKeyState(&TheTabState) && !TheTabState) {
			/* Advance the active Item and select NON button items */
			FormHighLight(TheDialog, FormPtr->CurrentActiveItem, False);
			FormPtr->CurrentActiveItem = FormPtr->CurrentActiveItem->NextItem;
			FormHighLight(TheDialog, FormPtr->CurrentActiveItem, True);
			if ((FormPtr->CurrentActiveItem->ItemType != FORM_DIALOG_BUTTON)   &&
			    (FormPtr->CurrentActiveItem->ItemType != FORM_FUNCTION_BUTTON) &&
			    (FormPtr->CurrentActiveItem->ItemType != FORM_SWITCH)
			   ) {
				FormPtr->CurrentItemIsSelected = True;
			}
		}
		else {
			/* Mouse Events */

			FormFoundOption = False;			
			FormItemPtr = FormPtr->ActiveItemList; 
			if (FormItemPtr != NULL) {
			    do {
				/* Loop over the Active Items */
				ItemX = (TheDialog->xpos+ FormItemPtr->DeltaX) - SH_WID;
				ItemY = (TheDialog->ypos+ FormItemPtr->DeltaY) - SH_WID;

				if (PointerInsideRegion(ItemX, ItemY, 
					FormItemPtr->width  + 2*SH_WID, 
					FormItemPtr->height + 2*SH_WID)) {
						
					FormFoundOption = True;
					break;
				}
				FormItemPtr = FormItemPtr->NextItem;
			    }
			    while (FormItemPtr != FormPtr->ActiveItemList); 
			}
			/* Select only when a Button Press event is detected */
			if ((PointerGetButtonState(&TheButtonState) & BUTTON_1) &&
				(!( TheButtonState & BUTTON_1))) {
				if (FormFoundOption) {
					FormPtr->CurrentItemIsSelected = True;
					FormHighLight(TheDialog, FormPtr->CurrentActiveItem, False);
					FormPtr->CurrentActiveItem = FormItemPtr;
					FormHighLight(TheDialog, FormPtr->CurrentActiveItem, True);
					FormPtr->CurrentItemIsSelected = True;

				}
/*
				else if (!( TheButtonState & BUTTON_1)) {
					if (TheDialog->ParentObject != NULL) {
						ActiveDialogObject = TheDialog->ParentObject;
						UnMapDialog(TheDialog);
					}
*/
				else if (TheDialog->ParentObject != NULL) {
						ActiveDialogObject = TheDialog->ParentObject;
						UnMapDialog(TheDialog);
				}
			}	

		}
	}
	if (KeyEvent == KEY_ESCAPE && TheDialog->ParentObject != NULL) {
		ActiveDialogObject = TheDialog->ParentObject;
		UnMapDialog(TheDialog);
	}

}
