
/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File menu.c : User interface for SimCoupe : Menu Object functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"

#include "dialog.h"

/* Pixmap used in Menus to indicate a submenu option */

static char MenuPixmapData[] = {SHCOL_DARK, SHCOL_DARK,   SHCOL_DARK,   SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_MEDIUM,
								SHCOL_DARK, SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_DARK,   SHCOL_DARK,   SHCOL_MEDIUM,
								SHCOL_DARK, SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_DARK,
								SHCOL_DARK, SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_LIGHT,
								SHCOL_DARK, SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_LIGHT,  SHCOL_LIGHT,  SHCOL_MEDIUM,								
								SHCOL_DARK, SHCOL_LIGHT,  SHCOL_LIGHT,  SHCOL_MEDIUM, SHCOL_MEDIUM, SHCOL_MEDIUM};

static Pixmap MenuPixmap = { 6, 6, MenuPixmapData };


/* Menu Object routines */

DialogObject *
MenuCreate(uns8 x, uns8 y, DialogObject *TheParent)
{
	/* Create a Menu. NB The actual size of a menu is determined   */
	/* dynamically when items are added			       */
	
	DialogObject *DialogPtr;
	
	DialogPtr = CreateDialogObject(MENU_OBJECT, x, y, 2*SH_WID, 2*SH_WID, TheParent);
	
	DialogPtr -> ObjectUnion.AMenu.NoMenuItems = 0;
	DialogPtr -> ObjectUnion.AMenu.CurrentItem = 0;
	return(DialogPtr);
	
}

void
MenuFree(DialogObject * TheDialog)
{
	/* Free up all the memory used by a menu, and the Free the Object itself */
	
	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	int Item;
	
	for (Item = 0; Item < MenuPtr -> NoMenuItems; Item++) {
		free(MenuPtr -> MenuItems[Item].ItemText);
	}
	FreeDialogObject(TheDialog);
}

int
MenuAddItem(DialogObject *TheDialog, int TheType, char *TheText, void* TheItemPtr)
{
	/* Add an item to the menu, adjusting the size appropriately */

	int EntryWidth;

	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	
	if (MenuPtr->NoMenuItems == MAX_MENU_ITEMS)
		return(False);

	MenuPtr->MenuItems[MenuPtr->NoMenuItems].ItemType = TheType;
	MenuPtr->MenuItems[MenuPtr->NoMenuItems].ItemText = (char *)malloc(1+strlen(TheText));
	strcpy(MenuPtr->MenuItems[MenuPtr->NoMenuItems].ItemText, TheText);
	MenuPtr->MenuItems[MenuPtr->NoMenuItems].ItemPtr  = TheItemPtr;	


	EntryWidth = strlen(TheText)*GLYPH_WIDTH;

	/* Submenu items include a pixmap, so we have to account for */
	/* the width of this objext as well.						 */
	if (TheType == MENU_SUBMENU_ITEM) {
		EntryWidth += MenuPixmap.PixmapWidth;
	}

	
	if (EntryWidth > TheDialog -> width) {
		 TheDialog -> width = 5*SH_WID + EntryWidth;
	}

	(TheDialog -> height) += GLYPH_HEIGHT + 2*SH_WID;

	(MenuPtr->NoMenuItems)++;

	return(True);
}

void
MenuDraw(DialogObject *TheDialog)
{
	uns8 ItemX, ItemY;
	int  Item;

	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);

	/* Draw the Menu using the appropriate primatives   	      */
	/* The Object should be mapped before this function is called */

	/* First we draw a shadow box */
	
	DrawShadowBox(TheDialog->xpos,  TheDialog->ypos,
		      TheDialog->width, TheDialog->height,
		      SHADOW_OUT, True, DM_OVER);

	/* Now write in the menu entries */

	ItemX = TheDialog->xpos+2*SH_WID;
	ItemY = TheDialog->ypos+2*SH_WID;	
	for (Item = 0; Item < MenuPtr->NoMenuItems; Item++) {
		DrawString(ItemX, ItemY, MenuPtr->MenuItems[Item].ItemText, SHCOL_BLACK, 
			   DM_OVER);

		/* For submenu entries we draw a pixmap */
		if (MenuPtr->MenuItems[Item].ItemType == MENU_SUBMENU_ITEM) {
			DrawPixmap((ItemX+TheDialog->width) - (4*SH_WID+MenuPixmap.PixmapWidth),
						ItemY+(GLYPH_HEIGHT-MenuPixmap.PixmapHeight)/2,
						&MenuPixmap, DM_OVER);		
		}
		ItemY += GLYPH_HEIGHT + 2*SH_WID;
		
	}	
}

void
MenuHighLight(DialogObject *TheDialog, int DoHighLight)
{
	/* Highlight or unhighlight a given menu option */
	
	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	uns8 ItemX = TheDialog->xpos+ SH_WID;
	uns8 ItemY = TheDialog->ypos+ SH_WID+ MenuPtr->CurrentItem*(GLYPH_HEIGHT + 2*SH_WID);	


	if (!DoHighLight)
		DrawBox(ItemX, ItemY, 
			TheDialog->width - 2*SH_WID, GLYPH_HEIGHT+2*SH_WID,
			SHCOL_MEDIUM, DM_OVER);
	else
		DrawShadowBox(ItemX, ItemY,
		      TheDialog->width - 2*SH_WID, GLYPH_HEIGHT+2*SH_WID,
		      SHADOW_OUT, False, DM_OVER);
 

}

void
MenuSelectItem(DialogObject *TheDialog, int TheItem)
{
	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	void (*ItemFunction)(void);

	/* Process a Menu selection */

	if (TheItem > MenuPtr->NoMenuItems ||
	    (MenuPtr -> MenuItems[TheItem].ItemPtr == NULL)) return;

	switch (MenuPtr -> MenuItems[TheItem].ItemType) {
	
		case MENU_DIALOG_ITEM: case MENU_SUBMENU_ITEM:
			ActiveDialogObject = (DialogObject *)MenuPtr->MenuItems[TheItem].ItemPtr;
			break;
			
		case MENU_FUNCTION_ITEM:
			ItemFunction = (void (*)(void))MenuPtr->MenuItems[TheItem].ItemPtr;
			ItemFunction();
			break;	
	
	}
}

void
MenuCore(DialogObject *TheDialog)
{
	MenuObject *MenuPtr = &(TheDialog->ObjectUnion.AMenu);
	int KeyEvent;
	uns8 ItemX, ItemY;
	int Item, MenuFoundOption, TheButtonState;


	/* The core menu code, called by DialogCore */
	
	/* Map and draw the menu if required */
	
	if (!TheDialog -> IsMapped) {
		MapDialog(TheDialog);
		MenuDraw(TheDialog);
		MenuHighLight(TheDialog, True);
	}

	/* Deal with Events */
	
	/* Keyboard events first */

/*	printf("CurrentItem = %d\n",MenuPtr->CurrentItem);*/
	switch (KeyEvent=GetSpecialKey()) {
	
		case KEY_UP:
			if (MenuPtr->CurrentItem) {
				MenuHighLight(TheDialog, False);
				MenuPtr->CurrentItem--;
				MenuHighLight(TheDialog, True);

			}
			break;
		case KEY_DOWN:
			if (MenuPtr->CurrentItem<MenuPtr->NoMenuItems-1) {
				MenuHighLight(TheDialog, False);
				MenuPtr->CurrentItem++;
				MenuHighLight(TheDialog, True);
			}
			break;	
		case KEY_ESCAPE: case KEY_LEFT:
			if (TheDialog->ParentObject != NULL) {
				ActiveDialogObject = TheDialog->ParentObject;
				UnMapDialog(TheDialog);
				return;
			}
			break;
		case KEY_RIGHT:
			if (MenuPtr->MenuItems[MenuPtr->CurrentItem].ItemType == MENU_SUBMENU_ITEM)
				MenuSelectItem(TheDialog, MenuPtr->CurrentItem);
			else if (TheDialog->ParentObject != NULL) {
				ActiveDialogObject = TheDialog->ParentObject;
				UnMapDialog(TheDialog);
				return;
			}
			break;

		case KEY_ENTER:
			MenuSelectItem(TheDialog, MenuPtr->CurrentItem);
			return;
			break;
	}

	/* Now the Mouse events */

	ItemX = TheDialog->xpos+2*SH_WID;
	ItemY = TheDialog->ypos+2*SH_WID;	
	MenuFoundOption = False;
	for (Item = 0; Item < MenuPtr->NoMenuItems; Item++) {
	
		if (PointerInsideRegion(ItemX, ItemY, 
					TheDialog->width - 2*SH_WID, 
					GLYPH_HEIGHT+2*SH_WID)) {
						
			MenuHighLight(TheDialog, False);
			MenuPtr->CurrentItem = Item;
			MenuHighLight(TheDialog, True);

			MenuFoundOption = True;

		}
		ItemY += GLYPH_HEIGHT+2*SH_WID;	
	}	

	if (PointerGetButtonState(&TheButtonState) & BUTTON_1) {
		if (MenuFoundOption) {
			MenuSelectItem(TheDialog, MenuPtr->CurrentItem);
		}
		else if (!( TheButtonState & BUTTON_1)) {
			if (TheDialog->ParentObject != NULL) {
				ActiveDialogObject = TheDialog->ParentObject;
				UnMapDialog(TheDialog);
			}
		}
	}	
}
