/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File pointer.c : User interface for SimCoupe : Mouse pointer functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"

#include "dialog.h"

/* default pointer Glyph and mask */

static char Glyphdata[] = {MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,0x00     ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,0x00     ,0x00     ,MASK_ALL ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,0x00     ,0x00     ,0x00     ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,0x00     ,0x00     ,0x00     ,0x00     ,MASK_ALL ,
			   MASK_ALL ,0x00     ,0x00     ,0x00     ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,MASK_ALL ,MASK_ALL ,0x00     ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,MASK_ALL ,MASK_ALL ,0x00     ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL };
			   
static char Maskdata[]  = {MASK_ALL ,MASK_ALL ,MASK_NONE,MASK_NONE,MASK_NONE,MASK_NONE,
			   MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_NONE,MASK_NONE,MASK_NONE,
			   MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL, MASK_NONE,MASK_NONE,
			   MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL, MASK_NONE,
			   MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_ALL ,
			   MASK_ALL ,MASK_ALL, MASK_ALL ,MASK_ALL ,MASK_ALL ,MASK_NONE,
			   MASK_NONE,MASK_NONE,MASK_ALL, MASK_ALL ,MASK_ALL ,MASK_NONE,
			   MASK_NONE,MASK_NONE,MASK_ALL, MASK_ALL ,MASK_ALL ,MASK_NONE};

static Pixmap DefaultGlyph = { 9, 6, Glyphdata };
static Pixmap DefaultMask  = { 9, 6, Maskdata  };
			       


/* Current Pointer */

static Pointer CurrentPointer;
static uns8 *PointerBackingStore = NULL;

void
PointerInit( void )
{
	/* Initialise the pointer */
	
	CurrentPointer.x = DISPLAY_WIDTH/2;
	CurrentPointer.y = DISPLAY_HEIGHT/2;
	CurrentPointer.ButtonState.ButtonEvent     = False;
	CurrentPointer.ButtonState.ButtonStatus    = 0x07;
        CurrentPointer.ButtonState.OldButtonStatus = 0x07;
        	
	CurrentPointer.Glyph = &DefaultGlyph;
	CurrentPointer.Mask  = &DefaultMask;

	PointerBackingStore = (uns8 *)malloc(CurrentPointer.Glyph->PixmapHeight*
					   CurrentPointer.Glyph->PixmapWidth);
}


void
PointerEnter( void )
{
	/* Fill the pointer backing store when the UI is entered */
	
	CurrentPointer.ButtonState.ButtonEvent     = False;
	CurrentPointer.ButtonState.ButtonStatus    = 0x07;
        CurrentPointer.ButtonState.OldButtonStatus = 0x07;
	PointerDraw();

}

void
PointerExit( void )
{
	/* Replace the pointer backing store when the UI is exited */
	
	PointerErase();

}

void 
PointerClose( void )
{


	/* Close down the pointer, freeing up the backing store */

	if (PointerBackingStore != NULL) {
		free (PointerBackingStore);	
		PointerBackingStore = NULL;
	}

}


void
PointerSetStatus( int TheDeltaX, int TheDeltaY, int TheButtonState )
{
	/* Set the Current Pointer status */

	int TestXY;

	CurrentPointer.x = (uns8)(TestXY = (int)CurrentPointer.x + TheDeltaX);

	if (TestXY < 0)			CurrentPointer.x = 0;
	if (TestXY > DISPLAY_WIDTH-1) 	CurrentPointer.x = DISPLAY_WIDTH-1;

	CurrentPointer.y = (uns8)(TestXY = (int)CurrentPointer.y + TheDeltaY);

	if (TestXY < 0)			CurrentPointer.y = 0;
	if (TestXY > DISPLAY_HEIGHT-1)	CurrentPointer.y = DISPLAY_HEIGHT-1;
	
	
	CurrentPointer.ButtonState.ButtonStatus = TheButtonState;
	CurrentPointer.ButtonState.ButtonEvent = (TheButtonState ^ CurrentPointer.ButtonState.OldButtonStatus) & 0x07;

#if 0
	simcoupe_mesg(SIMCOUPE_INFO,"Current : %d, Old : %d, Event : %d",
		CurrentPointer.ButtonState.ButtonStatus,
		CurrentPointer.ButtonState.OldButtonStatus,
		CurrentPointer.ButtonState.ButtonEvent
	);
#endif

	CurrentPointer.ButtonState.OldButtonStatus = TheButtonState;	
}

void
PointerSetGlyph( Pixmap *TheGlyph, Pixmap *TheMask )
{
	/* Set the Glyph for the current pointer */
	
	CurrentPointer.Glyph = TheGlyph;
	CurrentPointer.Mask  = TheMask;

	if (PointerBackingStore != NULL) {
		free (PointerBackingStore);
		PointerBackingStore = NULL;
	}

	PointerBackingStore = (uns8 *)malloc(CurrentPointer.Glyph->PixmapHeight*
					   CurrentPointer.Glyph->PixmapWidth);

	
}


Pointer *
PointerGetStatus( void )
{
	/* Get the current Pointer status */
	
	return &CurrentPointer;
	
}

uns8
PointerGetXPosition( void )
{
	/* Get the X position of the current pointer */
	
	return CurrentPointer.x;

}

uns8
PointerGetYPosition( void )
{
	/* Get the Y position of the current pointer */
	
	return CurrentPointer.y;

}


int 
PointerGetButtonState( int *TheStatus )
{
	/* Get the Button state of the current pointer */
	int TheEvent = CurrentPointer.ButtonState.ButtonEvent;
	
	*TheStatus  = CurrentPointer.ButtonState.ButtonStatus;

	return TheEvent;

}

int
PointerInsideRegion( uns8 TheXpos, uns8 TheYpos, uns8 TheWidth, uns8 TheHeight )
{
	/* Test whether the Current pointer is inside the specified region */

	
	if (CurrentPointer.x >= TheXpos &&
	    CurrentPointer.y >= TheYpos &&
	    CurrentPointer.x <  TheXpos+TheWidth &&
	    CurrentPointer.y <  TheYpos+TheHeight) 
	    
	    	return True;
	else
	
		return False;

}

void
PointerDraw( void )
{
	/* Draw the Current Pointer after saving what lies underneath */

/*	printf("x = %d, y=%d \n",CurrentPointer.x, CurrentPointer.y); */
	
	FetchArea(PointerBackingStore, CurrentPointer.x, CurrentPointer.y,
				       CurrentPointer.Glyph->PixmapWidth,
				       CurrentPointer.Glyph->PixmapHeight);
	 
	DrawPixmap(CurrentPointer.x, CurrentPointer.y, CurrentPointer.Mask, DM_OR);
	DrawPixmap(CurrentPointer.x, CurrentPointer.y, CurrentPointer.Glyph, DM_AND);
	
}

void
PointerErase( void )
{
	/* Erase the Current Pointer by writing over using the backing store */
	
	StoreArea(PointerBackingStore, CurrentPointer.x, CurrentPointer.y,
				       CurrentPointer.Glyph->PixmapWidth,
				       CurrentPointer.Glyph->PixmapHeight);

}
