/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File audio.c	- DOS interface to Aley Keprt's SAA emulator TSR
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */
 
#include <stdio.h> 
#include <string.h>
#include <stdlib.h>

#include <dpmi.h>
#include <sys/movedata.h>

#include "z80.h"
#include "sam.h"
#include "util.h"

#include "audio.h"

/* Values for %eax for int 0x21 calls */
#define TSR_CHECK	0x01126221
#define TSR_GET		0x01126223

/* Registers used for the int 0x21 call for TSR communication */
static  __dpmi_regs Int21Regs;

/* Real mode offset for TSR table */
	static uns32 TSROffset;

/* TSR info table structure */

typedef struct _TSRInfo {
	uns8 SAARegisters[NO_SAA_REGS];	/* SAA hardware registers     */
	uns32 pEntry;                	/* TSR entry point Seg:Offset */
	uns16 Version;			/* Current version            */
	uns16 OldVersion;		/* Oldest compatible version  */
	uns8  ChanMask;			/* Bit-mask of channels       */
	uns8  CurrentMode;		/* Current sound device mode  */
	uns8  MaxModes;			/* Number of sound dev. modes */
} TSRInfo;

static TSRInfo	AudioTSRTable;
static char    *AudioTSRFile;

int
AudioCheck(void)
{
	/* Initialize audio, check for TSR */

	static int AudioTSRInstalled = False;

	if (AudioTSRInstalled) {
		simcoupe_mesg(SIMCOUPE_WARN, "Audio TSR already installed");
		return False;
	}

	AudioTSRInstalled = True;

	Int21Regs.d.eax = TSR_CHECK;
	Int21Regs.x.bx  = 0x5b00;
	__dpmi_int(0x21, &Int21Regs);
  
	return (Int21Regs.x.bx == 1);
}

void
AudioTSRInitTable(void)
{
	/* Get a local copy of the TSR info table using int 0x21 */

	Int21Regs.d.eax = TSR_GET;
	Int21Regs.x.bx  = 0x5b00;
	__dpmi_int(0x21, &Int21Regs);

	TSROffset = (Int21Regs.x.es << 4) + Int21Regs.x.di;
 
	dosmemget(TSROffset,sizeof(TSRInfo), &AudioTSRTable);
}
uns8 *
AudioTSROperation(int AudioTSRFunction)
{
	/* Call the TSR player by using a simulated real mode    */
	/* call to the TSR entry point (found in the Info Table) */

	__dpmi_regs r;
	static char Buffer[256];

	r.x.ax = AudioTSRFunction;
	r.x.cs = AudioTSRTable.pEntry>>16;
	r.x.ip = AudioTSRTable.pEntry;
	r.x.ss = r.x.sp = 0x0;

	dosmemput(&AudioTSRTable, sizeof(TSRInfo), TSROffset);

	__dpmi_simulate_real_mode_procedure_retf(&r);

	/* now we have to copy returned string to local buffer */
	dosmemget((r.x.dx<<4)+r.x.ax,sizeof(Buffer),Buffer);
	return Buffer;
}

int
AudioLoadTSR(void)
{
	/* Install/remove audio TSR by executing the file */
	
	return (system(AudioTSRFile) == 0); 

}


uns8 *
AudioInit(char *TheTSRFile)
{
	/* API Init function Check and Install TSR table */
	/* Return pointer to SAA registers.              */

	/* Attempt to load the AudioTSR.                 */


	AudioTSRFile = TheTSRFile;
	AudioLoadTSR();

	if (!AudioCheck()) {
		simcoupe_mesg(SIMCOUPE_WARN, "No Audio TSR installed");
		return NULL;
	}

	simcoupe_mesg(SIMCOUPE_INFO, "Audio TSR installed, initializing...");

	AudioTSRInitTable();

	simcoupe_mesg(SIMCOUPE_INFO, "Version: %X.%02X",AudioTSRTable.Version / 256,
							AudioTSRTable.Version & 255);
	simcoupe_mesg(SIMCOUPE_INFO, "Lowest compatible Version: %X.%02X",
					AudioTSRTable.OldVersion / 256,
					AudioTSRTable.OldVersion & 255);

	simcoupe_mesg(SIMCOUPE_INFO, "Number of modes : %i", AudioTSRTable.MaxModes);

	simcoupe_mesg(SIMCOUPE_INFO, "TSR-player name : %s", 
					AudioTSROperation(AUDIO_TSR_NAME));
	simcoupe_mesg(SIMCOUPE_INFO, "Device Driver : %s", 
					AudioTSROperation(AUDIO_TSR_DEVICE));


	/* Clear the SAA registers */
	memset(AudioTSRTable.SAARegisters, 0x00, NO_SAA_REGS);

	/* Carry out an Open-Reset operation to fully reset player */

	AudioTSROperation(AUDIO_TSR_OPEN);
	AudioTSROperation(AUDIO_TSR_RESET);

	simcoupe_mesg(SIMCOUPE_INFO, "Audio TSR installation complete");

	return 	AudioTSRTable.SAARegisters;

}
void
AudioClose(void)
{

	/* Closedown the Audio TSR and remove the TSR driver */

	simcoupe_mesg(SIMCOUPE_INFO, "Closing down Audio TSR");
	AudioTSROperation(AUDIO_TSR_CLOSE);
	AudioLoadTSR();

}